# The complete language code info is from http://www.loc.gov/standards/iso639-2/php/code_list.php
# Language which has both "B" and "T" code will just use the "B" code in the list below
# code "nob", "pro","vol" have corresponding languages which include non-English character

languages = [
    ('aar', 'Afar'),
    ('abk', 'Abkhazian'),
    ('ace', 'Achinese'),
    ('ach', 'Acoli'),
    ('ada', 'Adangme'),
    ('ady', 'Adyghe; Adygei'),
    ('afa', 'Afro-Asiatic languages'),
    ('afh', 'Afrihili'),
    ('afr', 'Afrikaans'),
    ('ain', 'Ainu'),
    ('aka', 'Akan'),
    ('akk', 'Akkadian'),
    ('alb', 'Albanian'),
    ('ale', 'Aleut'),
    ('alg', 'Algonquian languages'),
    ('alt', 'Southern Altai'),
    ('amh', 'Amharic'),
    ('ang', 'English, Old (ca.450-1100)'),
    ('anp', 'Angika'),
    ('apa', 'Apache languages'),
    ('ara', 'Arabic'),
    ('arc', 'Official Aramaic (700-300 BCE); Imperial Aramaic (700-300 BCE)'),
    ('arg', 'Aragonese'),
    ('arm', 'Armenian'),
    ('arn', 'Mapudungun; Mapuche'),
    ('arp', 'Arapaho'),
    ('art', 'Artificial languages'),
    ('arw', 'Arawak'),
    ('asm', 'Assamese'),
    ('ast', 'Asturian; Bable; Leonese; Asturleonese'),
    ('ath', 'Athapascan languages'),
    ('aus', 'Australian languages'),
    ('ava', 'Avaric'),
    ('ave', 'Avestan'),
    ('awa', 'Awadhi'),
    ('aym', 'Aymara'),
    ('aze', 'Azerbaijani'),
    ('bad', 'Banda languages'),
    ('bai', 'Bamileke languages'),
    ('bak', 'Bashkir'),
    ('bal', 'Baluchi'),
    ('bam', 'Bambara'),
    ('ban', 'Balinese'),
    ('baq', 'Basque'),
    ('bas', 'Basa'),
    ('bat', 'Baltic languages'),
    ('bej', 'Beja; Bedawiyet'),
    ('bel', 'Belarusian'),
    ('bem', 'Bemba'),
    ('ben', 'Bengali'),
    ('ber', 'Berber languages'),
    ('bho', 'Bhojpuri'),
    ('bih', 'Bihari languages'),
    ('bik', 'Bikol'),
    ('bin', 'Bini; Edo'),
    ('bis', 'Bislama'),
    ('bla', 'Siksika'),
    ('bnt', 'Bantu languages'),
    ('tib', 'Tibetan'),
    ('bos', 'Bosnian'),
    ('bra', 'Braj'),
    ('bre', 'Breton'),
    ('btk', 'Batak languages'),
    ('bua', 'Buriat'),
    ('bug', 'Buginese'),
    ('bul', 'Bulgarian'),
    ('bur', 'Burmese'),
    ('byn', 'Blin; Bilin'),
    ('cad', 'Caddo'),
    ('cai', 'Central American Indian languages'),
    ('car', 'Galibi Carib'),
    ('cat', 'Catalan; Valencian'),
    ('cau', 'Caucasian languages'),
    ('ceb', 'Cebuano'),
    ('cel', 'Celtic languages'),
    ('cze', 'Czech'),
    ('cha', 'Chamorro'),
    ('chb', 'Chibcha'),
    ('che', 'Chechen'),
    ('chg', 'Chagatai'),
    ('chi', 'Chinese'),
    ('chk', 'Chuukese'),
    ('chm', 'Mari'),
    ('chn', 'Chinook jargon'),
    ('cho', 'Choctaw'),
    ('chp', 'Chipewyan; Dene Suline'),
    ('chr', 'Cherokee'),
    ('chu', 'Church Slavic; Old Slavonic; Church Slavonic; Old Bulgarian; Old Church Slavonic'),
    ('chv', 'Chuvash'),
    ('chy', 'Cheyenne'),
    ('cmc', 'Chamic languages'),
    ('cop', 'Coptic'),
    ('cor', 'Cornish'),
    ('cos', 'Corsican'),
    ('cpe', 'Creoles and pidgins, English based'),
    ('cpf', 'Creoles and pidgins, French-based'),
    ('cpp', 'Creoles and pidgins, Portuguese-based'),
    ('cre', 'Cree'),
    ('crh', 'Crimean Tatar; Crimean Turkish'),
    ('crp', 'Creoles and pidgins'),
    ('csb', 'Kashubian'),
    ('cus', 'Cushitic languages'),
    ('wel', 'Welsh'),
    ('dak', 'Dakota'),
    ('dan', 'Danish'),
    ('dar', 'Dargwa'),
    ('day', 'Land Dayak languages'),
    ('del', 'Delaware'),
    ('den', 'Slave (Athapascan)'),
    ('ger', 'German'),
    ('dgr', 'Dogrib'),
    ('din', 'Dinka'),
    ('div', 'Divehi; Dhivehi; Maldivian'),
    ('doi', 'Dogri'),
    ('dra', 'Dravidian languages'),
    ('dsb', 'Lower Sorbian'),
    ('dua', 'Duala'),
    ('dum', 'Dutch, Middle (ca.1050-1350)'),
    ('dut', 'Dutch; Flemish'),
    ('dyu', 'Dyula'),
    ('dzo', 'Dzongkha'),
    ('efi', 'Efik'),
    ('egy', 'Egyptian (Ancient)'),
    ('eka', 'Ekajuk'),
    ('gre', 'Greek, Modern (1453-)'),
    ('elx', 'Elamite'),
    ('eng', 'English'),
    ('enm', 'English, Middle (1100-1500)'),
    ('epo', 'Esperanto'),
    ('est', 'Estonian'),
    ('ewe', 'Ewe'),
    ('ewo', 'Ewondo'),
    ('fan', 'Fang'),
    ('fao', 'Faroese'),
    ('per', 'Persian'),
    ('fat', 'Fanti'),
    ('fij', 'Fijian'),
    ('fil', 'Filipino; Pilipino'),
    ('fin', 'Finnish'),
    ('fiu', 'Finno-Ugrian languages'),
    ('fon', 'Fon'),
    ('fre', 'French'),
    ('frm', 'French, Middle (ca.1400-1600)'),
    ('fro', 'French, Old (842-ca.1400)'),
    ('frr', 'Northern Frisian'),
    ('frs', 'Eastern Frisian'),
    ('fry', 'Western Frisian'),
    ('ful', 'Fulah'),
    ('fur', 'Friulian'),
    ('gaa', 'Ga'),
    ('gay', 'Gayo'),
    ('gba', 'Gbaya'),
    ('gem', 'Germanic languages'),
    ('geo', 'Georgian'),
    ('gez', 'Geez'),
    ('gil', 'Gilbertese'),
    ('gla', 'Gaelic; Scottish Gaelic'),
    ('gle', 'Irish'),
    ('glg', 'Galician'),
    ('glv', 'Manx'),
    ('gmh', 'German, Middle High (ca.1050-1500)'),
    ('goh', 'German, Old High (ca.750-1050)'),
    ('gon', 'Gondi'),
    ('gor', 'Gorontalo'),
    ('got', 'Gothic'),
    ('grb', 'Grebo'),
    ('grc', 'Greek, Ancient (to 1453)'),
    ('grn', 'Guarani'),
    ('gsw', 'Swiss German; Alemannic; Alsatian'),
    ('guj', 'Gujarati'),
    ('gwi', "Gwich'in"),
    ('hai', 'Haida'),
    ('hat', 'Haitian; Haitian Creole'),
    ('hau', 'Hausa'),
    ('haw', 'Hawaiian'),
    ('heb', 'Hebrew'),
    ('her', 'Herero'),
    ('hil', 'Hiligaynon'),
    ('him', 'Himachali languages; Western Pahari languages'),
    ('hin', 'Hindi'),
    ('hit', 'Hittite'),
    ('hmn', 'Hmong; Mong'),
    ('hmo', 'Hiri Motu'),
    ('hrv', 'Croatian'),
    ('hsb', 'Upper Sorbian'),
    ('hun', 'Hungarian'),
    ('hup', 'Hupa'),
    ('iba', 'Iban'),
    ('ibo', 'Igbo'),
    ('ice', 'Icelandic'),
    ('ido', 'Ido'),
    ('iii', 'Sichuan Yi; Nuosu'),
    ('ijo', 'Ijo languages'),
    ('iku', 'Inuktitut'),
    ('ile', 'Interlingue; Occidental'),
    ('ilo', 'Iloko'),
    ('ina', 'Interlingua (International Auxiliary Language Association)'),
    ('inc', 'Indic languages'),
    ('ind', 'Indonesian'),
    ('ine', 'Indo-European languages'),
    ('inh', 'Ingush'),
    ('ipk', 'Inupiaq'),
    ('ira', 'Iranian languages'),
    ('iro', 'Iroquoian languages'),
    ('ita', 'Italian'),
    ('jav', 'Javanese'),
    ('jbo', 'Lojban'),
    ('jpn', 'Japanese'),
    ('jpr', 'Judeo-Persian'),
    ('jrb', 'Judeo-Arabic'),
    ('kaa', 'Kara-Kalpak'),
    ('kab', 'Kabyle'),
    ('kac', 'Kachin; Jingpho'),
    ('kal', 'Kalaallisut; Greenlandic'),
    ('kam', 'Kamba'),
    ('kan', 'Kannada'),
    ('kar', 'Karen languages'),
    ('kas', 'Kashmiri'),
    ('kau', 'Kanuri'),
    ('kaw', 'Kawi'),
    ('kaz', 'Kazakh'),
    ('kbd', 'Kabardian'),
    ('kha', 'Khasi'),
    ('khi', 'Khoisan languages'),
    ('khm', 'Central Khmer'),
    ('kho', 'Khotanese; Sakan'),
    ('kik', 'Kikuyu; Gikuyu'),
    ('kin', 'Kinyarwanda'),
    ('kir', 'Kirghiz; Kyrgyz'),
    ('kmb', 'Kimbundu'),
    ('kok', 'Konkani'),
    ('kom', 'Komi'),
    ('kon', 'Kongo'),
    ('kor', 'Korean'),
    ('kos', 'Kosraean'),
    ('kpe', 'Kpelle'),
    ('krc', 'Karachay-Balkar'),
    ('krl', 'Karelian'),
    ('kro', 'Kru languages'),
    ('kru', 'Kurukh'),
    ('kua', 'Kuanyama; Kwanyama'),
    ('kum', 'Kumyk'),
    ('kur', 'Kurdish'),
    ('kut', 'Kutenai'),
    ('lad', 'Ladino'),
    ('lah', 'Lahnda'),
    ('lam', 'Lamba'),
    ('lao', 'Lao'),
    ('lat', 'Latin'),
    ('lav', 'Latvian'),
    ('lez', 'Lezghian'),
    ('lim', 'Limburgan; Limburger; Limburgish'),
    ('lin', 'Lingala'),
    ('lit', 'Lithuanian'),
    ('lol', 'Mongo'),
    ('loz', 'Lozi'),
    ('ltz', 'Luxembourgish; Letzeburgesch'),
    ('lua', 'Luba-Lulua'),
    ('lub', 'Luba-Katanga'),
    ('lug', 'Ganda'),
    ('lui', 'Luiseno'),
    ('lun', 'Lunda'),
    ('luo', 'Luo (Kenya and Tanzania)'),
    ('lus', 'Lushai'),
    ('mac', 'Macedonian'),
    ('mad', 'Madurese'),
    ('mag', 'Magahi'),
    ('mah', 'Marshallese'),
    ('mai', 'Maithili'),
    ('mak', 'Makasar'),
    ('mal', 'Malayalam'),
    ('man', 'Mandingo'),
    ('mao', 'Maori'),
    ('map', 'Austronesian languages'),
    ('mar', 'Marathi'),
    ('mas', 'Masai'),
    ('may', 'Malay'),
    ('mdf', 'Moksha'),
    ('mdr', 'Mandar'),
    ('men', 'Mende'),
    ('mga', 'Irish, Middle (900-1200)'),
    ('mic', "Mi'kmaq; Micmac"),
    ('min', 'Minangkabau'),
    ('mis', 'Uncoded languages'),
    ('mkh', 'Mon-Khmer languages'),
    ('mlg', 'Malagasy'),
    ('mlt', 'Maltese'),
    ('mnc', 'Manchu'),
    ('mni', 'Manipuri'),
    ('mno', 'Manobo languages'),
    ('moh', 'Mohawk'),
    ('mon', 'Mongolian'),
    ('mos', 'Mossi'),
    ('mul', 'Multiple languages'),
    ('mun', 'Munda languages'),
    ('mus', 'Creek'),
    ('mwl', 'Mirandese'),
    ('mwr', 'Marwari'),
    ('myn', 'Mayan languages'),
    ('myv', 'Erzya'),
    ('nah', 'Nahuatl languages'),
    ('nai', 'North American Indian languages'),
    ('nap', 'Neapolitan'),
    ('nau', 'Nauru'),
    ('nav', 'Navajo; Navaho'),
    ('nbl', 'Ndebele, South; South Ndebele'),
    ('nde', 'Ndebele, North; North Ndebele'),
    ('ndo', 'Ndonga'),
    ('nds', 'Low German; Low Saxon; German, Low; Saxon, Low'),
    ('nep', 'Nepali'),
    ('new', 'Nepal Bhasa; Newari'),
    ('nia', 'Nias'),
    ('nic', 'Niger-Kordofanian languages'),
    ('niu', 'Niuean'),
    ('nno', 'Norwegian Nynorsk; Nynorsk, Norwegian'),
    ('nob', 'Bokmal, Norwegian; Norwegian Bokmal'),
    ('nog', 'Nogai'),
    ('non', 'Norse, Old'),
    ('nor', 'Norwegian'),
    ('nqo', "N'Ko"),
    ('nso', 'Pedi; Sepedi; Northern Sotho'),
    ('nub', 'Nubian languages'),
    ('nwc', 'Classical Newari; Old Newari; Classical Nepal Bhasa'),
    ('nya', 'Chichewa; Chewa; Nyanja'),
    ('nym', 'Nyamwezi'),
    ('nyn', 'Nyankole'),
    ('nyo', 'Nyoro'),
    ('nzi', 'Nzima'),
    ('oci', 'Occitan (post 1500)'),
    ('oji', 'Ojibwa'),
    ('ori', 'Oriya'),
    ('orm', 'Oromo'),
    ('osa', 'Osage'),
    ('oss', 'Ossetian; Ossetic'),
    ('ota', 'Turkish, Ottoman (1500-1928)'),
    ('oto', 'Otomian languages'),
    ('paa', 'Papuan languages'),
    ('pag', 'Pangasinan'),
    ('pal', 'Pahlavi'),
    ('pam', 'Pampanga; Kapampangan'),
    ('pan', 'Panjabi; Punjabi'),
    ('pap', 'Papiamento'),
    ('pau', 'Palauan'),
    ('peo', 'Persian, Old (ca.600-400 B.C.)'),
    ('phi', 'Philippine languages'),
    ('phn', 'Phoenician'),
    ('pli', 'Pali'),
    ('pol', 'Polish'),
    ('pon', 'Pohnpeian'),
    ('por', 'Portuguese'),
    ('pra', 'Prakrit languages'),
    ('pro', 'Provencal, Old (to 1500);Occitan, Old (to 1500)'),
    ('pus', 'Pushto; Pashto'),
    ('qaa-qtz', 'Reserved for local use'),
    ('que', 'Quechua'),
    ('raj', 'Rajasthani'),
    ('rap', 'Rapanui'),
    ('rar', 'Rarotongan; Cook Islands Maori'),
    ('roa', 'Romance languages'),
    ('roh', 'Romansh'),
    ('rom', 'Romany'),
    ('rum', 'Romanian; Moldavian; Moldovan'),
    ('run', 'Rundi'),
    ('rup', 'Aromanian; Arumanian; Macedo-Romanian'),
    ('rus', 'Russian'),
    ('sad', 'Sandawe'),
    ('sag', 'Sango'),
    ('sah', 'Yakut'),
    ('sai', 'South American Indian languages'),
    ('sal', 'Salishan languages'),
    ('sam', 'Samaritan Aramaic'),
    ('san', 'Sanskrit'),
    ('sas', 'Sasak'),
    ('sat', 'Santali'),
    ('scn', 'Sicilian'),
    ('sco', 'Scots'),
    ('sel', 'Selkup'),
    ('sem', 'Semitic languages'),
    ('sga', 'Irish, Old (to 900)'),
    ('sgn', 'Sign Languages'),
    ('shn', 'Shan'),
    ('sid', 'Sidamo'),
    ('sin', 'Sinhala; Sinhalese'),
    ('sio', 'Siouan languages'),
    ('sit', 'Sino-Tibetan languages'),
    ('sla', 'Slavic languages'),
    ('slo', 'Slovak'),
    ('slv', 'Slovenian'),
    ('sma', 'Southern Sami'),
    ('sme', 'Northern Sami'),
    ('smi', 'Sami languages'),
    ('smj', 'Lule Sami'),
    ('smn', 'Inari Sami'),
    ('smo', 'Samoan'),
    ('sms', 'Skolt Sami'),
    ('sna', 'Shona'),
    ('snd', 'Sindhi'),
    ('snk', 'Soninke'),
    ('sog', 'Sogdian'),
    ('som', 'Somali'),
    ('son', 'Songhai languages'),
    ('sot', 'Sotho, Southern'),
    ('spa', 'Spanish; Castilian'),
    ('srd', 'Sardinian'),
    ('srn', 'Sranan Tongo'),
    ('srp', 'Serbian'),
    ('srr', 'Serer'),
    ('ssa', 'Nilo-Saharan languages'),
    ('ssw', 'Swati'),
    ('suk', 'Sukuma'),
    ('sun', 'Sundanese'),
    ('sus', 'Susu'),
    ('sux', 'Sumerian'),
    ('swa', 'Swahili'),
    ('swe', 'Swedish'),
    ('syc', 'Classical Syriac'),
    ('syr', 'Syriac'),
    ('tah', 'Tahitian'),
    ('tai', 'Tai languages'),
    ('tam', 'Tamil'),
    ('tat', 'Tatar'),
    ('tel', 'Telugu'),
    ('tem', 'Timne'),
    ('ter', 'Tereno'),
    ('tet', 'Tetum'),
    ('tgk', 'Tajik'),
    ('tgl', 'Tagalog'),
    ('tha', 'Thai'),
    ('tig', 'Tigre'),
    ('tir', 'Tigrinya'),
    ('tiv', 'Tiv'),
    ('tkl', 'Tokelau'),
    ('tlh', 'Klingon; tlhIngan-Hol'),
    ('tli', 'Tlingit'),
    ('tmh', 'Tamashek'),
    ('tog', 'Tonga (Nyasa)'),
    ('ton', 'Tonga (Tonga Islands)'),
    ('tpi', 'Tok Pisin'),
    ('tsi', 'Tsimshian'),
    ('tsn', 'Tswana'),
    ('tso', 'Tsonga'),
    ('tuk', 'Turkmen'),
    ('tum', 'Tumbuka'),
    ('tup', 'Tupi languages'),
    ('tur', 'Turkish'),
    ('tut', 'Altaic languages'),
    ('tvl', 'Tuvalu'),
    ('twi', 'Twi'),
    ('tyv', 'Tuvinian'),
    ('udm', 'Udmurt'),
    ('uga', 'Ugaritic'),
    ('uig', 'Uighur; Uyghur'),
    ('ukr', 'Ukrainian'),
    ('umb', 'Umbundu'),
    ('und', 'Undetermined'),
    ('urd', 'Urdu'),
    ('uzb', 'Uzbek'),
    ('vai', 'Vai'),
    ('ven', 'Venda'),
    ('vie', 'Vietnamese'),
    ('vol', 'Volapuk'),
    ('vot', 'Votic'),
    ('wak', 'Wakashan languages'),
    ('wal', 'Wolaitta; Wolaytta'),
    ('war', 'Waray'),
    ('was', 'Washo'),
    ('wen', 'Sorbian languages'),
    ('wln', 'Walloon'),
    ('wol', 'Wolof'),
    ('xal', 'Kalmyk; Oirat'),
    ('xho', 'Xhosa'),
    ('yao', 'Yao'),
    ('yap', 'Yapese'),
    ('yid', 'Yiddish'),
    ('yor', 'Yoruba'),
    ('ypk', 'Yupik languages'),
    ('zap', 'Zapotec'),
    ('zbl', 'Blissymbols; Blissymbolics; Bliss'),
    ('zen', 'Zenaga'),
    ('zgh', 'Standard Moroccan Tamazight'),
    ('zha', 'Zhuang; Chuang'),
    ('znd', 'Zande languages'),
    ('zul', 'Zulu'),
    ('zun', 'Zuni'),
    ('zxx', 'No linguistic content; Not applicable'),
    ('zza', 'Zaza; Dimili; Dimli; Kirdki; Kirmanjki; Zazaki'),
]



from rdflib import URIRef

from hsmodels.schemas.enums import CoverageType, DateType, MultidimensionalSpatialReferenceType, SpatialReferenceType
from hsmodels.utils import to_coverage_value_string


def parse_coverages(cls, values):
    if "coverages" in values:
        return values
    if "spatial_coverage" in values or "period_coverage" in values:
        assert "coverages" not in values
        values["coverages"] = []
    if "spatial_coverage" in values:
        sc = values["spatial_coverage"]
        del values["spatial_coverage"]
        if sc:
            value = to_coverage_value_string(sc)
            cov_type = CoverageType[sc["type"]]
            values["coverages"].append({"type": cov_type, "value": value})
    if "period_coverage" in values:
        pc = values["period_coverage"]
        del values["period_coverage"]
        if pc:
            value = to_coverage_value_string(pc)
            cov_type = CoverageType.period
            values["coverages"].append({"type": cov_type, "value": value})
    return values


def parse_rdf_spatial_reference(cls, values):
    if "spatial_reference" not in values:
        return values
    if not isinstance(values["spatial_reference"], dict):
        return values
    sr = values["spatial_reference"]
    value = to_coverage_value_string(sr)
    cov_type = SpatialReferenceType[sr["type"]]
    values["spatial_reference"] = {"type": cov_type, "value": value}
    return values


def parse_rdf_multidimensional_spatial_reference(cls, values):
    assert "spatial_reference" in values
    if not isinstance(values["spatial_reference"], dict):
        return values
    sr = values["spatial_reference"]
    value = to_coverage_value_string(sr)
    cov_type = MultidimensionalSpatialReferenceType[sr["type"]]
    values["spatial_reference"] = {"type": cov_type, "value": value}
    return values


def rdf_parse_description(cls, values):
    if "abstract" in values:
        values["description"] = {"abstract": values["abstract"]}
        del values["abstract"]
    return values


def rdf_parse_utc_offset(cls, values):
    if "value" in values:
        values["utc_offset"] = {"value": values["value"]}
        del values["value"]
    return values


def rdf_parse_rdf_subject(cls, values):
    if "url" in values:
        values["rdf_subject"] = URIRef(values["url"])
        del values["url"]
    return values


def parse_rdf_extended_metadata(cls, values):
    if "additional_metadata" in values:
        em = values["additional_metadata"]
        assert isinstance(em, dict)
        values["extended_metadata"] = []
        del values["additional_metadata"]
        for key, value in em.items():
            values["extended_metadata"].append({"key": key, "value": value})
    return values


def parse_rdf_dates(cls, values):
    if "dates" in values:
        return values

    dates = []
    assert "created" in values
    assert values["created"]
    dates.append({"type": DateType.created, "value": values["created"]})
    del values["created"]

    assert "modified" in values
    assert values["modified"]
    dates.append({"type": DateType.modified, "value": values["modified"]})
    del values["modified"]

    if "published" in values and values["published"]:
        dates.append({"type": DateType.published, "value": values["published"]})
        del values["published"]

    values["dates"] = dates
    return values


def parse_relation_rdf(cls, values):
    if "type" in values and "value" in values:
        values[values["type"].name] = values["value"]
    return values


def split_user_identifiers(cls, values):
    if "identifiers" in values:
        identifiers = values["identifiers"]
        for id_type, id_value in identifiers.items():
            values[id_type.name] = id_value
    return values



from datetime import datetime
from typing import Any, Dict, List, Type, Union

from pydantic import AnyUrl, Field, root_validator, validator

from hsmodels.schemas.base_models import BaseMetadata
from hsmodels.schemas.fields import (
    AwardInfo,
    BoxCoverage,
    Contributor,
    Creator,
    PeriodCoverage,
    PointCoverage,
    Publisher,
    Relation,
    Rights,
)
from hsmodels.schemas.rdf.validators import language_constraint
from hsmodels.schemas.root_validators import (
    parse_abstract,
    parse_additional_metadata,
    parse_url,
    split_coverages,
    split_dates,
)
from hsmodels.schemas.validators import list_not_empty, parse_identifier, parse_sources, parse_spatial_coverage


class ResourceMetadata(BaseMetadata):
    """
    A class used to represent the metadata for a resource
    """

    class Config:
        title = 'Resource Metadata'

        schema_config = {'read_only': ['type', 'identifier', 'created', 'modified', 'published', 'url']}

    type: str = Field(
        const=True,
        default="CompositeResource",
        title="Resource Type",
        description="An object containing a URL that points to the HydroShare resource type selected from the hsterms namespace",
        allow_mutation=False,
    )

    url: AnyUrl = Field(title="URL", description="An object containing the URL for a resource", allow_mutation=False)

    identifier: AnyUrl = Field(
        title="Identifier",
        description="An object containing the URL-encoded unique identifier for a resource",
        allow_mutation=False,
    )
    title: str = Field(
        max_length=300, default=None, title="Title", description="A string containing the name given to a resource"
    )
    abstract: str = Field(default=None, title="Abstract", description="A string containing a summary of a resource")
    language: str = Field(
        title="Language",
        description="A 3-character string for the language in which the metadata and content of a resource are expressed",
    )
    subjects: List[str] = Field(
        default=[], title="Subject keywords", description="A list of keyword strings expressing the topic of a resource"
    )
    creators: List[Creator] = Field(
        default=[],
        title="Creators",
        description="A list of Creator objects indicating the entities responsible for creating a resource",
    )
    contributors: List[Contributor] = Field(
        default=[],
        title="Contributors",
        description="A list of Contributor objects indicating the entities that contributed to a resource",
    )
    sources: List[str] = Field(
        default=[],
        title="Sources",
        description="A list of strings containing references to related resources from which a described resource was derived",
    )
    relations: List[Relation] = Field(
        default=[],
        title="Related resources",
        description="A list of Relation objects representing resources related to a described resource",
    )
    additional_metadata: Dict[str, str] = Field(
        default={},
        title="Additional metadata",
        description="A dictionary containing key-value pair metadata associated with a resource",
    )
    rights: Rights = Field(
        title="Rights", description="An object containing information about rights held in an over a resource"
    )
    created: datetime = Field(
        default_factory=datetime.now,
        title="Creation date",
        description="A datetime object containing the instant associated with when a resource was created",
        allow_mutation=False,
    )
    modified: datetime = Field(
        default_factory=datetime.now,
        title="Modified date",
        description="A datetime object containing the instant associated with when a resource was last modified",
        allow_mutation=False,
    )
    published: datetime = Field(
        default=None,
        title="Published date",
        description="A datetime object containing the instant associated with when a resource was published",
        allow_mutation=False,
    )
    awards: List[AwardInfo] = Field(
        default=[],
        title="Funding agency information",
        description="A list of objects containing information about the funding agencies and awards associated with a resource",
    )
    spatial_coverage: Union[PointCoverage, BoxCoverage] = Field(
        default=None,
        title="Spatial coverage",
        description="An object containing information about the spatial topic of a resource, the spatial applicability of a resource, or jurisdiction under with a resource is relevant",
    )
    period_coverage: PeriodCoverage = Field(
        default=None,
        title="Temporal coverage",
        description="An object containing information about the temporal topic or applicability of a resource",
    )
    publisher: Publisher = Field(
        default=None,
        title="Publisher",
        description="An object containing information about the publisher of a resource",
    )
    citation: str = Field(
        default=None, title="Citation", description="A string containing the biblilographic citation for a resource"
    )

    _parse_coverages = root_validator(pre=True, allow_reuse=True)(split_coverages)
    _parse_dates = root_validator(pre=True, allow_reuse=True)(split_dates)
    _parse_additional_metadata = root_validator(pre=True, allow_reuse=True)(parse_additional_metadata)
    _parse_abstract = root_validator(pre=True)(parse_abstract)
    _parse_url = root_validator(pre=True, allow_reuse=True)(parse_url)

    _parse_identifier = validator("identifier", pre=True)(parse_identifier)
    _parse_sources = validator("sources", pre=True)(parse_sources)
    _parse_spatial_coverage = validator("spatial_coverage", pre=True)(parse_spatial_coverage)

    _language_constraint = validator('language', allow_reuse=True)(language_constraint)
    _creators_constraint = validator('creators')(list_not_empty)



from typing import List

from pydantic import AnyUrl, Field, root_validator

from hsmodels.namespaces import DC, HSTERMS, RDF
from hsmodels.schemas.rdf.fields import (
    BandInformationInRDF,
    CellInformationInRDF,
    CoverageInRDF,
    DescriptionInRDF,
    ExtendedMetadataInRDF,
    FieldInformationInRDF,
    GeometryInformationInRDF,
    MultidimensionalSpatialReferenceInRDF,
    RDFBaseModel,
    RightsInRDF,
    SpatialReferenceInRDF,
    TimeSeriesResultInRDF,
    VariableInRDF,
)
from hsmodels.schemas.rdf.root_validators import (
    parse_coverages,
    parse_rdf_extended_metadata,
    parse_rdf_multidimensional_spatial_reference,
    parse_rdf_spatial_reference,
    rdf_parse_description,
    rdf_parse_rdf_subject,
)


class BaseAggregationMetadataInRDF(RDFBaseModel):
    _parse_rdf_subject = root_validator(pre=True, allow_reuse=True)(rdf_parse_rdf_subject)

    title: str = Field(rdf_predicate=DC.title)
    subjects: List[str] = Field(rdf_predicate=DC.subject, default=[])
    language: str = Field(rdf_predicate=DC.language, default="eng")
    extended_metadata: List[ExtendedMetadataInRDF] = Field(rdf_predicate=HSTERMS.extendedMetadata, default=[])
    coverages: List[CoverageInRDF] = Field(rdf_predicate=DC.coverage, default=[])
    rights: RightsInRDF = Field(rdf_predicate=DC.rights, default=[])

    _parse_coverages = root_validator(pre=True, allow_reuse=True)(parse_coverages)

    _parse_extended_metadata = root_validator(pre=True, allow_reuse=True)(parse_rdf_extended_metadata)


class GeographicRasterMetadataInRDF(BaseAggregationMetadataInRDF):
    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=HSTERMS.GeographicRasterAggregation)

    label: str = Field(
        const=True,
        default="Geographic Raster Content: A geographic grid represented by a virtual "
        "raster tile (.vrt) file and one or more geotiff (.tif) files",
    )
    dc_type: AnyUrl = Field(rdf_predicate=DC.type, default=HSTERMS.GeographicRasterAggregation, const=True)

    band_information: BandInformationInRDF = Field(rdf_predicate=HSTERMS.BandInformation)
    spatial_reference: SpatialReferenceInRDF = Field(rdf_predicate=HSTERMS.spatialReference, default=None)
    cell_information: CellInformationInRDF = Field(rdf_predicate=HSTERMS.CellInformation)

    _parse_spatial_reference = root_validator(pre=True, allow_reuse=True)(parse_rdf_spatial_reference)


class GeographicFeatureMetadataInRDF(BaseAggregationMetadataInRDF):
    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=HSTERMS.GeographicFeatureAggregation)

    label: str = Field(
        const=True, default="Geographic Feature Content: The multiple files that are part of a " "geographic shapefile"
    )
    dc_type: AnyUrl = Field(rdf_predicate=DC.type, default=HSTERMS.GeographicFeatureAggregation, const=True)

    field_information: List[FieldInformationInRDF] = Field(rdf_predicate=HSTERMS.FieldInformation)
    geometry_information: GeometryInformationInRDF = Field(rdf_predicate=HSTERMS.GeometryInformation)
    spatial_reference: SpatialReferenceInRDF = Field(rdf_predicate=HSTERMS.spatialReference, default=None)

    _parse_spatial_reference = root_validator(pre=True, allow_reuse=True)(parse_rdf_spatial_reference)


class MultidimensionalMetadataInRDF(BaseAggregationMetadataInRDF):
    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=HSTERMS.MultidimensionalAggregation)

    label: str = Field(
        const=True,
        default="Multidimensional Content: A multidimensional dataset represented by a "
        "NetCDF file (.nc) and text file giving its NetCDF header content",
    )
    dc_type: AnyUrl = Field(rdf_predicate=DC.type, default=HSTERMS.MultidimensionalAggregation, const=True)

    variables: List[VariableInRDF] = Field(rdf_predicate=HSTERMS.Variable)
    spatial_reference: MultidimensionalSpatialReferenceInRDF = Field(
        rdf_predicate=HSTERMS.spatialReference, default=None
    )

    _parse_spatial_reference = root_validator(pre=True, allow_reuse=True)(parse_rdf_multidimensional_spatial_reference)


class TimeSeriesMetadataInRDF(BaseAggregationMetadataInRDF):
    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=HSTERMS.TimeSeriesAggregation)

    label: str = Field(
        const=True,
        default="Time Series Content: One or more time series held in an ODM2 format "
        "SQLite file and optional source comma separated (.csv) files",
    )
    dc_type: AnyUrl = Field(rdf_predicate=DC.type, default=HSTERMS.TimeSeriesAggregation, const=True)
    description: DescriptionInRDF = Field(rdf_predicate=DC.description, default_factory=DescriptionInRDF)

    time_series_results: List[TimeSeriesResultInRDF] = Field(rdf_predicate=HSTERMS.timeSeriesResult)

    _parse_description = root_validator(pre=True, allow_reuse=True)(rdf_parse_description)


class ReferencedTimeSeriesMetadataInRDF(BaseAggregationMetadataInRDF):
    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=HSTERMS.ReferencedTimeSeriesAggregation)

    label: str = Field(
        const=True,
        default="Referenced Time Series Content: A reference to one or more time series "
        "served from HydroServers outside of HydroShare in WaterML format",
    )
    dc_type: AnyUrl = Field(rdf_predicate=DC.type, default=HSTERMS.ReferencedTimeSeriesAggregation, const=True)


class FileSetMetadataInRDF(BaseAggregationMetadataInRDF):
    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=HSTERMS.FileSetAggregation)

    label: str = Field(const=True, default="File Set Content: One or more files with specific metadata")
    dc_type: AnyUrl = Field(rdf_predicate=DC.type, default=HSTERMS.FileSetAggregation, const=True)


class SingleFileMetadataInRDF(BaseAggregationMetadataInRDF):
    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=HSTERMS.SingleFileAggregation)

    label: str = Field(const=True, default="Single File Content: A single file with file specific metadata")
    dc_type: AnyUrl = Field(rdf_predicate=DC.type, default=HSTERMS.SingleFileAggregation, const=True)



from datetime import datetime
from typing import Dict

from pydantic import AnyUrl, EmailStr, Field, HttpUrl, root_validator, validator

from hsmodels.schemas import base_models
from hsmodels.schemas.base_models import BaseMetadata
from hsmodels.schemas.enums import RelationType, UserIdentifierType, VariableType
from hsmodels.schemas.root_validators import group_user_identifiers, parse_relation, parse_utc_offset_value
from hsmodels.schemas.validators import validate_user_url


class Relation(BaseMetadata):
    """
    A class used to represent the metadata associated with a resource related to the resource being described
    """

    class Config:
        title = 'Related Resource Metadata'

    type: RelationType = Field(title="Relation type", description="The type of relationship with the related resource")
    value: str = Field(
        max_length=500,
        title="Value",
        description="String expressing the Full text citation, URL link for, or description of the related resource",
    )

    _parse_relation = root_validator(pre=True)(parse_relation)


class CellInformation(BaseMetadata):
    """
    A class used to represent the metadata associated with raster grid cells in geographic raster aggregations
    """

    class Config:
        title = 'Raster Cell Metadata'

    # TODO: Is there such a thing as "name" for CellInformation?
    name: str = Field(default=None, max_length=500, title="Name", description="Name of the cell information")
    rows: int = Field(default=None, title="Rows", description="The integer number of rows in the raster dataset")
    columns: int = Field(
        default=None, title="Columns", description="The integer number of columns in the raster dataset"
    )
    cell_size_x_value: float = Field(
        default=None,
        title="Cell size x value",
        description="The size of the raster grid cell in the x-direction expressed as a float",
    )
    cell_data_type: str = Field(
        default=None, max_length=50, title="Cell data type", description="The data type of the raster grid cell values"
    )
    cell_size_y_value: float = Field(
        default=None,
        title="Cell size y value",
        description="The size of the raster grid cell in the y-direction expressed as a float",
    )


class Rights(BaseMetadata):
    """
    A class used to represent the rights statement metadata associated with a resource
    """

    class Config:
        title = 'Rights Metadata'

    statement: str = Field(
        title="Statement", description="A string containing the text of the license or rights statement"
    )
    url: AnyUrl = Field(
        title="URL",
        description="An object containing the URL pointing to a description of the license or rights statement",
    )

    @classmethod
    def Creative_Commons_Attribution_CC_BY(cls):
        return Rights(
            statement="This resource is shared under the Creative Commons Attribution CC BY.",
            url="http://creativecommons.org/licenses/by/4.0/",
        )

    @classmethod
    def Creative_Commons_Attribution_ShareAlike_CC_BY(cls):
        return Rights(
            statement="This resource is shared under the Creative Commons Attribution-ShareAlike CC BY-SA.",
            url="http://creativecommons.org/licenses/by-sa/4.0/",
        )

    @classmethod
    def Creative_Commons_Attribution_NoDerivs_CC_BY_ND(cls):
        return Rights(
            statement="This resource is shared under the Creative Commons Attribution-ShareAlike CC BY-SA.",
            url="http://creativecommons.org/licenses/by-nd/4.0/",
        )

    @classmethod
    def Creative_Commons_Attribution_NoCommercial_ShareAlike_CC_BY_NC_SA(cls):
        return Rights(
            statement="This resource is shared under the Creative Commons Attribution-NoCommercial-ShareAlike"
            " CC BY-NC-SA.",
            url="http://creativecommons.org/licenses/by-nc-sa/4.0/",
        )

    @classmethod
    def Creative_Commons_Attribution_NoCommercial_CC_BY_NC(cls):
        return Rights(
            statement="This resource is shared under the Creative Commons Attribution-NoCommercial CC BY-NC.",
            url="http://creativecommons.org/licenses/by-nc/4.0/",
        )

    @classmethod
    def Creative_Commons_Attribution_NoCommercial_NoDerivs_CC_BY_NC_ND(cls):
        return Rights(
            statement="This resource is shared under the Creative Commons Attribution-NoCommercial-NoDerivs "
            "CC BY-NC-ND.",
            url="http://creativecommons.org/licenses/by-nc-nd/4.0/",
        )

    @classmethod
    def Other(cls, statement: str, url: AnyUrl):
        return Rights(statement=statement, url=url)


class Creator(BaseMetadata):
    """
    A class used to represent the metadata associated with a creator of a resource
    """

    class Config:
        title = 'Creator Metadata'

        schema_config = {'read_only': ['description']}

    name: str = Field(
        default=None, max_length=100, title="Name", description="A string containing the name of the creator"
    )
    phone: str = Field(
        default=None, max_length=25, title="Phone", description="A string containing a phone number for the creator"
    )
    address: str = Field(
        default=None, max_length=250, title="Address", description="A string containing an address for the creator"
    )
    organization: str = Field(
        default=None,
        max_length=200,
        title="Organization",
        description="A string containing the name of the organization with which the creator is affiliated",
    )
    email: EmailStr = Field(
        default=None, title="Email", description="A string containing an email address for the creator"
    )
    homepage: HttpUrl = Field(
        default=None,
        title="Homepage",
        description="An object containing the URL for website associated with the creator",
    )
    # TODO: Is there such a thing as a "description" property for Creator?
    description: str = Field(
        max_length=50,
        default=None,
        title="Description",
        description="A string containing the path to the hydroshare profile",
        allow_mutation=False,
    )
    identifiers: Dict[UserIdentifierType, AnyUrl] = Field(
        default={},
        title="Creator identifiers",
        description="A dictionary containing identifier types and URL links to alternative identifiers for the creator",
    )

    _description_validator = validator("description", pre=True)(validate_user_url)

    _split_identifiers = root_validator(pre=True, allow_reuse=True)(group_user_identifiers)

    @classmethod
    def from_user(cls, user):
        user_dict = user.dict()
        user_dict["description"] = user.url.path
        if user.website:
            user_dict["homepage"] = user.website

        return Creator(**user_dict)


class Contributor(BaseMetadata):
    """
    A class used to represent the metadata associated with a contributor to a resource
    """

    class Config:
        title = 'Contributor Metadata'

        schema_config = {'read_only': ['description']}

    name: str = Field(default=None, title="Name", description="A string containing the name of the contributor")
    phone: str = Field(
        default=None, title="Phone", description="A string containing a phone number for the contributor"
    )
    address: str = Field(
        default=None, title="Address", description="A string containing an address for the contributor"
    )
    organization: str = Field(
        default=None,
        title="Organization",
        description="A string containing the name of the organization with which the contributor is affiliated",
    )
    email: EmailStr = Field(
        default=None, title="Email", description="A string containing an email address for the contributor"
    )
    homepage: HttpUrl = Field(
        default=None,
        title="Homepage",
        description="An object containing the URL for website associated with the contributor",
    )
    description: str = Field(
        max_length=50,
        default=None,
        title="Description",
        description="A string containing the path to the hydroshare profile",
        allow_mutation=False,
    )
    identifiers: Dict[UserIdentifierType, AnyUrl] = Field(
        default={},
        title="Contributor identifiers",
        description="A dictionary containing identifier types and URL links to alternative identiers for the contributor",
    )

    _split_identifiers = root_validator(pre=True, allow_reuse=True)(group_user_identifiers)

    @classmethod
    def from_user(cls, user):
        """
        Constructs a Contributor from a User object
        :param user: a User
        :return: a Contributor
        """
        user_dict = user.dict()
        user_dict["description"] = user.url.path
        if user.website:
            user_dict["homepage"] = user.website

        return Contributor(**user_dict)


class AwardInfo(BaseMetadata):
    """
    A class used to represent the metadata associated with funding agency credits for a resource
    """

    class Config:
        title = 'Funding Agency Metadata'

    funding_agency_name: str = Field(
        title="Agency name", description="A string containing the name of the funding agency or organization"
    )
    title: str = Field(
        default=None, title="Award title", description="A string containing the title of the project or award"
    )
    number: str = Field(
        default=None, title="Award number", description="A string containing the award number or other identifier"
    )
    funding_agency_url: AnyUrl = Field(
        default=None,
        title="Agency URL",
        description="An object containing a URL pointing to a website describing the funding award",
    )


class BandInformation(BaseMetadata):
    """
    A class used to represent the metadata associated with the raster bands of a geographic raster aggregation
    """

    class Config:
        title = 'Raster Band Metadata'

    name: str = Field(max_length=500, title="Name", description="A string containing the name of the raster band")
    variable_name: str = Field(
        default=None,
        max_length=100,
        title="Variable name",
        description="A string containing the name of the variable represented by the raster band",
    )
    variable_unit: str = Field(
        default=None,
        max_length=50,
        title="Variable unit",
        description="A string containing the units for the raster band variable",
    )
    no_data_value: str = Field(
        default=None,
        title="Nodata value",
        description="A string containing the numeric nodata value for the raster band",
    )
    maximum_value: str = Field(
        default=None,
        title="Maximum value",
        description="A string containing the maximum numeric value for the raster band",
    )
    comment: str = Field(
        default=None, title="Comment", description="A string containing a comment about the raster band"
    )
    method: str = Field(
        default=None,
        title="Method",
        description="A string containing a description of the method used to create the raster band data",
    )
    minimum_value: str = Field(
        default=None,
        title="Minimum value",
        description="A string containing the minimum numerica value for the raster dataset",
    )


class FieldInformation(BaseMetadata):
    """
    A class used to represent the metadata associated with a field in the attribute table for a geographic
    feature aggregation
    """

    class Config:
        title = 'Geographic Feature Field Metadata'

    field_name: str = Field(
        max_length=128, title="Field name", description="A string containing the name of the attribute table field"
    )
    field_type: str = Field(
        max_length=128, title="Field type", description="A string containing the data type of the values in the field"
    )
    # TODO: What is the "field_type_code"? It's not displayed on the resource landing page, but it's encoded in the
    #  aggregation metadata as an integer value.
    field_type_code: str = Field(
        default=None,
        max_length=50,
        title="Field type code",
        description="A string value containing a code that indicates the field type",
    )
    field_width: int = Field(
        default=None, title="Field width", description="An integer value containing the width of the attribute field"
    )
    field_precision: int = Field(
        default=None,
        title="Field precision",
        description="An integer value containing the precision of the attribute field",
    )


class GeometryInformation(BaseMetadata):
    """
    A class used to represent the metadata associated with the geometry of a geographic feature aggregation
    """

    class Config:
        title = 'Geographic Feature Geometry Metadata'

    feature_count: int = Field(
        default=0,
        title="Feature count",
        description="An integer containing the number of features in the geographic feature aggregation",
    )
    geometry_type: str = Field(
        max_length=128,
        title="Geometry type",
        description="A string containing the type of features in the geographic feature aggregation",
    )


class Variable(BaseMetadata):
    """
    A class used to represent the metadata associated with a variable contained within a multidimensional aggregation
    """

    class Config:
        title = 'Multidimensional Variable Metadata'

    name: str = Field(
        max_length=1000, title="Variable name", description="A string containing the name of the variable"
    )
    unit: str = Field(
        max_length=1000,
        title="Units",
        description="A string containing the units in which the values for the variable are expressed",
    )
    type: VariableType = Field(title="Type", description="The data type of the values for the variable")
    shape: str = Field(
        max_length=1000,
        title="Shape",
        description="A string containing the shape of the variable expressed as a list of dimensions",
    )
    descriptive_name: str = Field(
        default=None,
        max_length=1000,
        title="Descriptive name",
        description="A string containing a descriptive name for the variable",
    )
    method: str = Field(
        default=None,
        title="Method",
        description="A string containing a description of the method used to create the values for the variable",
    )
    missing_value: str = Field(
        default=None,
        max_length=1000,
        title="Missing value",
        description="A string containing the value used to indicate missing values for the variable",
    )


class Publisher(BaseMetadata):
    """
    A class used to represent the metadata associated with the publisher of a resource
    """

    class Config:
        title = 'Publisher Metadata'

    name: str = Field(
        max_length=200, title="Publisher name", description="A string containing the name of the publisher"
    )
    url: AnyUrl = Field(
        title="Publisher URL", description="An object containing a URL that points to the publisher website"
    )


class TimeSeriesVariable(BaseMetadata):
    """
    A class used to represent the metadata associated with a variable contained within a time series aggregation
    """

    class Config:
        title = 'Time Series Variable Metadata'

    variable_code: str = Field(
        max_length=50,
        title="Variable code",
        description="A string containing a short but meaningful code that identifies a variable",
    )
    variable_name: str = Field(
        max_length=100, title="Variable name", description="A string containing the name of the variable"
    )
    variable_type: str = Field(
        max_length=100,
        title="Variable type",
        description="A string containing the type of variable from the ODM2 VariableType controlled vocabulary",
    )
    # TODO: The NoData value for a variable in an ODM2 database is not always an integer.
    #  It could be a floating point value. We might want to change this to a string or a floating point value
    #  It is an integer in the HydroShare database, so will have to be updated there as well if changed
    no_data_value: int = Field(title="NoData value", description="The NoData value for the variable")
    variable_definition: str = Field(
        default=None,
        max_length=255,
        title="Variable definition",
        description="A string containing a detailed description of the variable",
    )
    speciation: str = Field(
        default=None,
        max_length=255,
        title="Speciation",
        description="A string containing the speciation for the variable from the ODM2 Speciation controllec vocabulary",
    )


class TimeSeriesSite(BaseMetadata):
    """
    A class used to represent the metadata associated with a site contained within a time series aggregation
    """

    class Config:
        title = 'Time Series Site Metadata'

    site_code: str = Field(
        max_length=200,
        title="Site code",
        description="A string containing a short but meaningful code identifying the site",
    )
    site_name: str = Field(
        default=None, max_length=255, title="Site name", description="A string containing the name of the site"
    )
    elevation_m: float = Field(
        default=None,
        title="Elevation",
        description="A floating point number expressing the elevation of the site in meters",
    )
    elevation_datum: str = Field(
        default=None,
        max_length=50,
        title="Elevation datum",
        description="A string expressing the elevation datum used from the ODM2 Elevation Datum controlled vocabulary",
    )
    site_type: str = Field(
        default=None,
        max_length=100,
        title="Site type",
        description="A string containing the type of site from the ODM2 Sampling Feature Type controlled vocabulary ",
    )
    latitude: float = Field(
        default=None,
        title="Latitude",
        description="A floating point value expressing the latitude coordinate of the site",
    )
    longitude: float = Field(
        default=None,
        title="Longitude",
        description="A floating point value expressing the longitude coordinate of the site",
    )


class TimeSeriesMethod(BaseMetadata):
    """
    A class used to represent the metadata associated with a method contained within a time series aggregation
    """

    class Config:
        title = 'Time Series Method Metadata'

    method_code: str = Field(
        max_length=50,
        title="Method code",
        description="A string containing a short but meaningful code identifying the method",
    )
    method_name: str = Field(
        max_length=200, title="Method name", description="A string containing the name of the method"
    )
    method_type: str = Field(
        max_length=200,
        title="Method type",
        description="A string containing the method type from the ODM2 Method Type controlled vocabulary",
    )
    method_description: str = Field(
        default=None, title="Method description", description="A string containing a detailed description of the method"
    )
    method_link: AnyUrl = Field(
        default=None,
        title="Method link",
        description="An object containg a URL that points to a website having a detailed description of the method",
    )


class ProcessingLevel(BaseMetadata):
    """
    A class used to represent the metadata associated with a processing level contained within a time series
    aggregation
    """

    class Config:
        title = 'Time Series Processing Level Metadata'

    processing_level_code: str = Field(
        max_length=50,
        title="Processing level code",
        description="A string containing a short but meaningful code identifying the processing level",
    )
    definition: str = Field(
        default=None,
        max_length=200,
        title="Definition",
        description="A string containing a description of the processing level",
    )
    explanation: str = Field(
        default=None,
        title="Explanation",
        description="A string containing a more extensive explanation of the meaning of the processing level",
    )


class Unit(BaseMetadata):
    """
    A class used to represent the metadata associated with a dimensional unit within a time series aggregation
    """

    class Config:
        title = 'Time Series Units Metadata'

    type: str = Field(
        max_length=255,
        title="Unit type",
        description="A string containing the type of unit from the ODM2 Units Type controlled vocabulary",
    )
    name: str = Field(
        max_length=255,
        title="Unit name",
        description="A string containing the name of the unit from the ODM2 units list",
    )
    abbreviation: str = Field(
        max_length=20,
        title="Unit abbreviation",
        description="A string containing an abbreviation for the unit from the ODM2 units list",
    )


class UTCOffSet(BaseMetadata):
    """
    A class used to represent the metadata associated with a UTC time offset within a time series aggregation)
    """

    class Config:
        title = 'Time Series UTC Offset Metadata'

    value: float = Field(
        default=0,
        title="UTC offset value",
        description="A floating point number containing the UTC time offset associated with the data values expressed in hours",
    )


class TimeSeriesResult(BaseMetadata):
    """
    A class used to represent the metadata associated with a time series result within a time series aggregation
    """

    class Config:
        title = 'Time Series Result Metadata'

    series_id: str = Field(
        max_length=36,
        title="Series ID",
        description="A string containing a unique identifier for the time series result",
    )
    unit: Unit = Field(
        default=None,
        title="Units",
        description="An object containing the units in which the values of the time series are expressed",
    )
    status: str = Field(
        default=None,
        max_length=255,
        title="Status",
        description="A string containing the status of the time series result chosen from the ODM2 Status controlled vocabulary",
    )
    sample_medium: str = Field(
        max_length=255,
        title="Sample medium",
        description="A string containing the sample medium in which the time series result was measured chosen from the ODM2 Medium controlled vocabulary",
    )
    value_count: int = Field(
        title="Value count",
        description="An integer value containing the number of data values contained within the time series result",
    )
    aggregation_statistic: str = Field(
        max_length=255,
        title="Aggregation statistic",
        description="A string containing the aggregation statistic associated with the values of the time series result chosen from the ODM2 Aggregation Statistic controlled vocabulary",
    )
    # TODO: Not sure what "series_label" is. It's not an ODM2 thing
    # in HydroShare it is generated with this format
    # label = "{site_code}:{site_name}, {variable_code}:{variable_name}, {units_name}, {pro_level_code}, {method_name}"
    series_label: str = Field(
        default=None,
        max_length=255,
        title="Series label",
        description="A string containing a label for the time series result",
    )
    site: TimeSeriesSite = Field(
        title="Site",
        description="An object containing metadata about the site at which the time series result was created",
    )
    variable: TimeSeriesVariable = Field(
        title="Variablef",
        description="An object containing metadata about the observed variable associated with the time series result values",
    )
    method: TimeSeriesMethod = Field(
        title="Method",
        description="An object containing metadata about the method used to produce the time series result values",
    )
    processing_level: ProcessingLevel = Field(
        title="Processing level",
        description="An object containing metadata about the processing level or level of quality control to which the time series result values have been subjected",
    )
    utc_offset: float = Field(
        default=None,
        title="UTC Offset",
        description="A floating point value that represents the time offset from UTC time in hours associated with the time series result value timestamps",
    )
    _parse_utc_offset = root_validator(pre=True, allow_reuse=True)(parse_utc_offset_value)


class BoxCoverage(base_models.BaseCoverage):
    """
    A class used to represent geographic coverage metadata for a resource or aggregation expressed as a
    latitude-longitude bounding box
    """

    class Config:
        title = 'Box Coverage Metadata'

        schema_config = {'read_only': ['type']}

    type: str = Field(
        default="box",
        const=True,
        title="Geographic coverage type",
        description="A string containing the type of geographic coverage",
        allow_mutation=False,
    )
    name: str = Field(
        default=None,
        title="Name",
        description="A string containing a name for the place associated with the geographic coverage",
    )
    northlimit: float = Field(
        gt=-90,
        lt=90,
        title="North limit",
        description="A floating point value containing the constant coordinate for the northernmost face or edge of the bounding box",
    )
    eastlimit: float = Field(
        gt=-180,
        lt=180,
        title="East limit",
        description="A floating point value containing the constant coordinate for the easternmost face or edge of the bounding box",
    )
    southlimit: float = Field(
        gt=-90,
        lt=90,
        title="South limit",
        description="A floating point value containing the constant coordinate for the southernmost face or edge of the bounding box",
    )
    westlimit: float = Field(
        gt=-180,
        lt=180,
        title="West limit",
        description="A floating point value containing the constant coordinate for the westernmost face or edge of the bounding box",
    )
    units: str = Field(
        title="Units",
        description="A string containing the units applying to the unlabelled numeric values of northlimit, eastlimit, southlimit, and westlimit",
    )
    projection: str = Field(
        default=None,
        title="Projection",
        description="A string containing the name of the projection used with any parameters required, such as ellipsoid parameters, datum, standard parallels and meridians, zone, etc.",
    )

    @root_validator
    def compare_north_south(cls, values):
        north, south = values["northlimit"], values["southlimit"]
        if north < south:
            raise ValueError(f"North latitude [{north}] must be greater than or equal to South latitude [{south}]")
        return values


class BoxSpatialReference(base_models.BaseCoverage):
    """
    A class used to represent the metadata associated with the spatial reference of a geographic
    feature or raster aggregation expressed as a bounding box
    """

    class Config:
        title = 'Box Spatial Reference Metadata'

        schema_config = {'read_only': ['type']}

    type: str = Field(
        default="box",
        const=True,
        title="Spatial reference type",
        description="A string containing the type of spatial reference",
        allow_mutation=False,
    )
    name: str = Field(
        default=None,
        title="Name",
        description="A string containing a name for the place associated with the spatial reference",
    )
    northlimit: float = Field(
        title="North limit",
        description="A floating point value containing the constant coordinate for the northernmost face or edge of the bounding box",
    )
    eastlimit: float = Field(
        title="East limit",
        description="A floating point value containing the constant coordinate for the easternmost face or edge of the bounding box",
    )
    southlimit: float = Field(
        title="South limit",
        description="A floating point value containing the constant coordinate for the southernmost face or edge of the bounding box",
    )
    westlimit: float = Field(
        title="West limit",
        description="A floating point value containing the constant coordinate for the westernmost face or edge of the bounding box",
    )
    units: str = Field(
        title="Units",
        description="A string containing the units applying to the unlabelled numeric values of northlimit, eastlimit, southlimit, and westlimit",
    )
    # TODO: "projection" should probably be "projection_name"
    projection: str = Field(
        default=None,
        title="Projection",
        description="A string containing the name of the coordinate system used by the spatial reference",
    )
    projection_string: str = Field(
        title="Projection string", description="A string containing an encoding of the coordinate system parameters"
    )
    # TODO: I'm not sure what "projection_string_type" is - I don't see it in the RDF/XML encoding
    projection_string_type: str = Field(
        default=None,
        title="Projection string type",
        description="A string containing a description of the type of encoding for the projection string",
    )
    datum: str = Field(
        default=None,
        title="Datum",
        description="A string containing the name of the datum used by the coordinate system",
    )
    projection_name: str = Field(
        default=None, title="Projection name", description="A string containing the name of the coordinate system"
    )


class MultidimensionalBoxSpatialReference(BoxSpatialReference):
    """
    A class used to represent the metadata associated with the spatial reference of a multidimensional
    aggregation expressed as a bounding box
    """

    class Config:
        title = 'Multidimensional Box Spatial Reference Metadata'


class PointCoverage(base_models.BaseCoverage):
    """
    A class used to represent geographic coverage metadata for a resource or aggregation expressed as a
    point location
    """

    class Config:
        title = 'Point Coverage Metadata'

        schema_config = {'read_only': ['type']}

    type: str = Field(
        default="point",
        const=True,
        title="Geographic coverage type",
        description="A string containing the type of geographic coverage",
        allow_mutation=False,
    )
    name: str = Field(
        default=None,
        title="Name",
        description="A string containing a name for the place associated with the geographic coverage",
    )
    east: float = Field(
        gt=-180, lt=180, title="East", description="The coordinate of the point location measured in the east direction"
    )
    north: float = Field(
        gt=-90, lt=90, title="North", description="The coordinate of the point location measured in the north direction"
    )
    units: str = Field(
        title="Units", description="The units applying to the unlabelled numeric values of north and east"
    )
    projection: str = Field(
        title="Projection",
        description="The name of the projection used with any parameters required, such as ellipsoid parameters, datum, standard parallels and meridians, zone, etc.",
    )


class PointSpatialReference(base_models.BaseCoverage):
    """
    A class used to represent the metadata associated with the spatial reference of a geographic
    feature or raster aggregation expressed as a point
    """

    class Config:
        title = 'Point Spatial Reference Metadata'

        schema_config = {'read_only': ['type']}

    type: str = Field(
        default="point",
        const=True,
        title="Spatial reference type",
        description="A string containing the type of spatial reference",
        allow_mutation=False,
    )
    name: str = Field(
        default=None,
        title="Name",
        description="A string containing a name for the place associated with the spatial reference",
    )
    east: float = Field(title="East", description="The coordinate of the point location measured in the east direction")
    north: float = Field(
        title="North", description="The coordinate of the point location measured in the north direction"
    )
    units: str = Field(
        title="Units", description="The units applying to the unlabelled numeric values of north and east"
    )
    projection: str = Field(
        title="Projection",
        description="A string containing the name of the coordinate system used by the spatial reference",
    )
    projection_string: str = Field(
        title="Projection string", description="A string containing an encoding of the coordinate system parameters"
    )
    # TODO: I'm not sure what "projection_string_type" is - I don't see it in the RDF/XML encoding
    projection_string_type: str = Field(
        default=None,
        title="Projection string type",
        description="A string containing a description of the type of encoding for the projection string",
    )
    projection_name: str = Field(
        default=None, title="Projection name", description="A string containing the name of the coordinate system"
    )


class MultidimensionalPointSpatialReference(PointSpatialReference):
    """
    A class used to represent the metadata associated with the spatial reference of a multidimensional
    aggregation expressed as a point
    """

    class Config:
        title = 'Multidimensional Point Spatial Reference Metadata'


class PeriodCoverage(base_models.BaseCoverage):
    """
    A class used to represent temporal coverage metadata for a resource or aggregation
    """

    class Config:
        title = 'Period Coverage Metadata'

    name: str = Field(default=None, title="Name", description="A string containing a name for the time interval")
    start: datetime = Field(
        title="Start",
        description="A datetime object containing the instant corresponding to the commencement of the time interval",
    )
    end: datetime = Field(
        title="End",
        description="A datetime object containing the instant corresponding to the termination of the time interval",
    )

    @root_validator
    def start_before_end(cls, values):
        if "start" in values and "end" in values:
            start, end = values["start"], values["end"]
            if start > end:
                raise ValueError(f"start date [{start}] is after end date [{end}]")
        return values



import inspect
from datetime import datetime
from enum import Enum

from pydantic import AnyUrl, BaseModel
from rdflib import Graph, Literal, URIRef

from hsmodels.namespaces import DC, HSTERMS, ORE, RDF, RDFS1, XSD
from hsmodels.schemas.aggregations import (
    FileSetMetadata,
    GeographicFeatureMetadata,
    GeographicRasterMetadata,
    MultidimensionalMetadata,
    ReferencedTimeSeriesMetadata,
    SingleFileMetadata,
    TimeSeriesMetadata,
)
from hsmodels.schemas.enums import TermEnum
from hsmodels.schemas.rdf.aggregations import (
    FileSetMetadataInRDF,
    GeographicFeatureMetadataInRDF,
    GeographicRasterMetadataInRDF,
    MultidimensionalMetadataInRDF,
    ReferencedTimeSeriesMetadataInRDF,
    SingleFileMetadataInRDF,
    TimeSeriesMetadataInRDF,
)
from hsmodels.schemas.rdf.resource import ResourceMap, ResourceMetadataInRDF
from hsmodels.schemas.resource import ResourceMetadata

rdf_schemas = {
    ORE.ResourceMap: ResourceMap,
    HSTERMS.CompositeResource: ResourceMetadataInRDF,
    HSTERMS.GeographicRasterAggregation: GeographicRasterMetadataInRDF,
    HSTERMS.GeographicFeatureAggregation: GeographicFeatureMetadataInRDF,
    HSTERMS.MultidimensionalAggregation: MultidimensionalMetadataInRDF,
    HSTERMS.ReferencedTimeSeriesAggregation: ReferencedTimeSeriesMetadataInRDF,
    HSTERMS.FileSetAggregation: FileSetMetadataInRDF,
    HSTERMS.SingleFileAggregation: SingleFileMetadataInRDF,
    HSTERMS.TimeSeriesAggregation: TimeSeriesMetadataInRDF,
}

user_schemas = {
    ResourceMetadataInRDF: ResourceMetadata,
    GeographicRasterMetadataInRDF: GeographicRasterMetadata,
    GeographicFeatureMetadataInRDF: GeographicFeatureMetadata,
    MultidimensionalMetadataInRDF: MultidimensionalMetadata,
    ReferencedTimeSeriesMetadataInRDF: ReferencedTimeSeriesMetadata,
    FileSetMetadataInRDF: FileSetMetadata,
    SingleFileMetadataInRDF: SingleFileMetadata,
    TimeSeriesMetadataInRDF: TimeSeriesMetadata,
}


def load_rdf(rdf_str, file_format='xml'):

    g = Graph().parse(data=rdf_str, format=file_format)
    for target_class, schema in rdf_schemas.items():
        subject = g.value(predicate=RDF.type, object=target_class)
        if subject:
            if target_class == ResourceMap:
                return _parse(schema, g)
            else:
                rdf_metadata = _parse(schema, g)
                if schema in user_schemas.keys():
                    return user_schemas[schema](**rdf_metadata.dict())
                return rdf_metadata
    raise Exception("Could not find schema for \n{}".format(rdf_str))


def parse_file(schema, file, file_format='xml', subject=None):
    metadata_graph = Graph().parse(file, format=file_format)
    return _parse(schema, metadata_graph, subject)


def rdf_graph(schema):
    for rdf_schema, user_schema in user_schemas.items():
        if isinstance(schema, user_schema):
            return _rdf_graph(rdf_schema(**schema.dict()), Graph())
    return _rdf_graph(schema, Graph())


def rdf_string(schema, rdf_format='pretty-xml'):
    return rdf_graph(schema).serialize(format=rdf_format).decode()


def _rdf_fields(schema):
    for f in schema.__fields__.values():
        if f.alias not in ['rdf_subject', 'rdf_type', 'label', 'dc_type']:
            predicate = f.field_info.extra.get('rdf_predicate', None)
            if not predicate:
                config_field_info = schema.Config.fields.get(f.name, None)
                if isinstance(config_field_info, dict):
                    predicate = config_field_info.get('rdf_predicate', None)
            if not predicate:
                raise Exception(
                    "Schema configuration error for {}, all fields must specify a rdf_predicate".format(schema)
                )
            yield f, predicate


def _rdf_graph(schema, graph=None):
    for f, predicate in _rdf_fields(schema):
        values = getattr(schema, f.name, None)
        if values:
            if not isinstance(values, list):
                # handle single values as a list to simplify
                values = [values]
            for value in values:
                if isinstance(value, BaseModel):
                    assert hasattr(value, "rdf_subject")
                    # nested class
                    graph.add((schema.rdf_subject, predicate, value.rdf_subject))
                    graph = _rdf_graph(value, graph)
                else:
                    # primitive value
                    if isinstance(value, AnyUrl):
                        value = URIRef(value)
                    elif isinstance(value, datetime):
                        value = Literal(value.isoformat())
                    elif isinstance(value, int):
                        value = Literal(value, datatype=XSD.integer)
                    elif isinstance(value, float):
                        value = Literal(value, datatype=XSD.double)
                    elif isinstance(value, TermEnum):
                        value = URIRef(value.value)
                    elif isinstance(value, Enum):
                        value = Literal(value.value)
                    else:
                        value = Literal(value)
                    graph.add((schema.rdf_subject, predicate, value))
    if hasattr(schema, 'rdf_type'):
        graph.add((schema.rdf_subject, RDF.type, schema.rdf_type))
    if hasattr(schema, 'dc_type'):
        graph.add((schema.rdf_subject, DC.type, schema.dc_type))
    if hasattr(schema, 'label'):
        graph.add((URIRef(str(schema.rdf_type)), RDFS1.label, Literal(schema.label)))
        graph.add((URIRef(str(schema.rdf_type)), RDFS1.isDefinedBy, URIRef("https://www.hydroshare.org/terms/")))
    return graph


def get_args(t):
    return getattr(t, "__args__", None)


def _parse(schema, metadata_graph, subject=None):
    def nested_class(field):
        if field.sub_fields:
            clazz = get_args(field.outer_type_)[0]
        else:
            clazz = field.outer_type_
        if inspect.isclass(clazz):
            return issubclass(clazz, BaseModel)
        return False

    def class_rdf_type(schema):
        if schema.__fields__['rdf_type']:
            return schema.__fields__['rdf_type'].default
        return None

    if not subject:
        # lookup subject using RDF.type specified in the schema
        target_class = class_rdf_type(schema)
        if not target_class:
            raise Exception("Subject must be provided, no RDF.type specified on class {}".format(schema))
        subject = metadata_graph.value(predicate=RDF.type, object=target_class)
        if not subject:
            raise Exception("Could not find subject for predicate=RDF.type, object={}".format(target_class))
    kwargs = {}
    for f, predicate in _rdf_fields(schema):
        parsed = []
        for value in metadata_graph.objects(subject=subject, predicate=predicate):
            if nested_class(f):
                if f.sub_fields:
                    # list
                    clazz = f.sub_fields[0].outer_type_
                else:
                    # single
                    clazz = f.outer_type_
                parsed_class = _parse(clazz, metadata_graph, value)
                if parsed_class:
                    parsed.append(parsed_class)
                elif f.sub_fields:
                    parsed.append([])
            else:
                # primitive value
                parsed.append(str(value.toPython()))
        if len(parsed) > 0:
            if f.sub_fields:
                # list
                kwargs[f.name] = parsed
            else:
                # single
                kwargs[f.name] = parsed[0]
    if kwargs:
        instance = schema(**kwargs, rdf_subject=subject)
        return instance
    return None



from rdflib import Namespace

HSRESOURCE = Namespace("http://www.hydroshare.org/resource/")
DCTERMS = Namespace("http://purl.org/dc/terms/")
RDFS1 = Namespace("http://www.w3.org/2000/01/rdf-schema#")
SCHEMA = Namespace("http://schema.org/")
HSTERMS = Namespace("https://www.hydroshare.org/terms/")
XML = Namespace("http://www.w3.org/XML/1998/namespace")
RDFS = Namespace("http://www.w3.org/2000/01/rdf-schema#")
DC = Namespace("http://purl.org/dc/elements/1.1/")
CITOTERMS = Namespace("http://purl.org/spar/cito/")
XSD = Namespace("http://www.w3.org/2001/XMLSchema#")
SH = Namespace("http://www.w3.org/ns/shacl#")
RDF = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
FOAF = Namespace("http://xmlns.com/foaf/0.1/")
DASH = Namespace("http://datashapes.org/dash#")
ORE = Namespace("http://www.openarchives.org/ore/terms/")
HSUSER = Namespace("http://www.hydroshare.org/user/")



from datetime import datetime

from pydantic import AnyUrl, BaseModel, EmailStr, Field, HttpUrl, PositiveInt, root_validator
from rdflib import BNode
from rdflib.term import Identifier as RDFIdentifier

from hsmodels.namespaces import DCTERMS, HSTERMS, RDF, RDFS
from hsmodels.schemas.enums import CoverageType, DateType, MultidimensionalSpatialReferenceType, SpatialReferenceType
from hsmodels.schemas.fields import (
    AwardInfo,
    BandInformation,
    CellInformation,
    FieldInformation,
    GeometryInformation,
    ProcessingLevel,
    Publisher,
    Rights,
    TimeSeriesMethod,
    TimeSeriesResult,
    TimeSeriesSite,
    TimeSeriesVariable,
    Unit,
    UTCOffSet,
    Variable,
)
from hsmodels.schemas.rdf.root_validators import parse_relation_rdf, rdf_parse_utc_offset, split_user_identifiers


class RDFBaseModel(BaseModel):
    rdf_subject: RDFIdentifier = Field(default_factory=BNode)


class DCTypeInRDF(RDFBaseModel):
    is_defined_by: AnyUrl = Field(rdf_predicate=RDFS.isDefinedBy)
    label: str = Field(rdf_predicate=RDFS.label)


class SourceInRDF(RDFBaseModel):
    is_derived_from: str = Field(rdf_predicate=HSTERMS.isDerivedFrom, default=None)


class RelationInRDF(RDFBaseModel):
    isHostedBy: str = Field(rdf_predicate=HSTERMS.isHostedBy, default=None)
    isCopiedFrom: str = Field(rdf_predicate=HSTERMS.isCopiedFrom, default=None)
    isPartOf: str = Field(rdf_predicate=HSTERMS.isPartOf, default=None)
    hasPart: str = Field(rdf_predicate=HSTERMS.hasPart, default=None)
    isExecutedBy: str = Field(rdf_predicate=HSTERMS.isExecutedBy, default=None)
    isCreatedBy: str = Field(rdf_predicate=HSTERMS.isCreatedBy, default=None)
    isVersionOf: str = Field(rdf_predicate=HSTERMS.isVersionOf, default=None)
    isReplacedBy: str = Field(rdf_predicate=HSTERMS.isReplacedBy, default=None)
    isDataFor: str = Field(rdf_predicate=HSTERMS.isDataFor, default=None)
    cites: str = Field(rdf_predicate=HSTERMS.cites, default=None)
    isDescribedBy: str = Field(rdf_predicate=HSTERMS.isDescribedBy, default=None)

    _parse_relation = root_validator(pre=True)(parse_relation_rdf)


class DescriptionInRDF(RDFBaseModel):
    abstract: str = Field(rdf_predicate=DCTERMS.abstract, default=None)


class IdentifierInRDF(RDFBaseModel):
    hydroshare_identifier: AnyUrl = Field(rdf_predicate=HSTERMS.hydroShareIdentifier)


class ExtendedMetadataInRDF(RDFBaseModel):
    value: str = Field(rdf_predicate=HSTERMS.value)
    key: str = Field(rdf_predicate=HSTERMS.key)


class CellInformationInRDF(CellInformation, RDFBaseModel):
    class Config:
        fields = {
            'name': {"rdf_predicate": HSTERMS.name},
            'rows': {"rdf_predicate": HSTERMS.rows},
            'columns': {"rdf_predicate": HSTERMS.columns},
            'cell_size_x_value': {"rdf_predicate": HSTERMS.cellSizeXValue},
            'cell_data_type': {"rdf_predicate": HSTERMS.cellDataType},
            'cell_size_y_value': {"rdf_predicate": HSTERMS.cellSizeYValue},
        }


class DateInRDF(RDFBaseModel):
    type: DateType = Field(rdf_predicate=RDF.type)
    value: datetime = Field(rdf_predicate=RDF.value)


class RightsInRDF(Rights, RDFBaseModel):
    class Config:
        fields = {'statement': {"rdf_predicate": HSTERMS.rightsStatement}, 'url': {"rdf_predicate": HSTERMS.URL}}


class CreatorInRDF(RDFBaseModel):
    creator_order: PositiveInt
    name: str = Field(default=None)
    phone: str = Field(default=None)
    address: str = Field(default=None)
    organization: str = Field(default=None)
    email: EmailStr = Field(default=None)
    homepage: HttpUrl = Field(default=None)
    description: str = Field(max_length=50, default=None)
    ORCID: AnyUrl = Field(default=None)
    google_scholar_id: AnyUrl = Field(default=None)
    research_gate_id: AnyUrl = Field(default=None)

    _group_identifiers = root_validator(pre=True, allow_reuse=True)(split_user_identifiers)

    class Config:
        fields = {
            'name': {"rdf_predicate": HSTERMS.name},
            'creator_order': {"rdf_predicate": HSTERMS.creatorOrder},
            'google_scholar_id': {"rdf_predicate": HSTERMS.GoogleScholarID},
            'research_gate_id': {"rdf_predicate": HSTERMS.ResearchGateID},
            'phone': {"rdf_predicate": HSTERMS.phone},
            'ORCID': {"rdf_predicate": HSTERMS.ORCID},
            'address': {"rdf_predicate": HSTERMS.address},
            'organization': {"rdf_predicate": HSTERMS.organization},
            'email': {"rdf_predicate": HSTERMS.email},
            'homepage': {"rdf_predicate": HSTERMS.homepage},
            'description': {"rdf_predicate": HSTERMS.description},
        }


class ContributorInRDF(RDFBaseModel):
    name: str = Field(default=None)
    phone: str = Field(default=None)
    address: str = Field(default=None)
    organization: str = Field(default=None)
    email: EmailStr = Field(default=None)
    homepage: HttpUrl = Field(default=None)
    description: str = Field(max_length=50, default=None)
    ORCID: AnyUrl = Field(default=None)
    google_scholar_id: AnyUrl = Field(default=None)
    research_gate_id: AnyUrl = Field(default=None)

    _group_identifiers = root_validator(pre=True, allow_reuse=True)(split_user_identifiers)

    class Config:
        fields = {
            'name': {"rdf_predicate": HSTERMS.name},
            'phone': {"rdf_predicate": HSTERMS.phone},
            'address': {"rdf_predicate": HSTERMS.address},
            'organization': {"rdf_predicate": HSTERMS.organization},
            'email': {"rdf_predicate": HSTERMS.email},
            'homepage': {"rdf_predicate": HSTERMS.homepage},
            'ORCID': {"rdf_predicate": HSTERMS.ORCID},
            'google_scholar_id': {"rdf_predicate": HSTERMS.GoogleScholarID},
            'research_gate_id': {"rdf_predicate": HSTERMS.ResearchGateID},
            'description': {"rdf_predicate": HSTERMS.description},
        }


class AwardInfoInRDF(AwardInfo, RDFBaseModel):
    class Config:
        fields = {
            'funding_agency_name': {"rdf_predicate": HSTERMS.fundingAgencyName},
            'title': {"rdf_predicate": HSTERMS.awardTitle},
            'number': {"rdf_predicate": HSTERMS.awardNumber},
            'funding_agency_url': {"rdf_predicate": HSTERMS.fundingAgencyURL},
        }


class BandInformationInRDF(BandInformation, RDFBaseModel):
    class Config:
        fields = {
            'name': {"rdf_predicate": HSTERMS.name},
            'variable_name': {"rdf_predicate": HSTERMS.variableName},
            'variable_unit': {"rdf_predicate": HSTERMS.variableUnit},
            'no_data_value': {"rdf_predicate": HSTERMS.noDataValue},
            'maximum_value': {"rdf_predicate": HSTERMS.maximumValue},
            'comment': {"rdf_predicate": HSTERMS.comment},
            'method': {"rdf_predicate": HSTERMS.method},
            'minimum_value': {"rdf_predicate": HSTERMS.minimumValue},
        }


class CoverageInRDF(RDFBaseModel):
    type: CoverageType = Field(rdf_predicate=RDF.type)
    value: str = Field(rdf_predicate=RDF.value)


class SpatialReferenceInRDF(RDFBaseModel):
    type: SpatialReferenceType = Field(rdf_predicate=RDF.type)
    value: str = Field(rdf_predicate=RDF.value)


class MultidimensionalSpatialReferenceInRDF(RDFBaseModel):
    type: MultidimensionalSpatialReferenceType = Field(rdf_predicate=RDF.type)
    value: str = Field(rdf_predicate=RDF.value)


class FieldInformationInRDF(FieldInformation, RDFBaseModel):
    class Config:
        fields = {
            'field_name': {"rdf_predicate": HSTERMS.fieldName},
            'field_type': {"rdf_predicate": HSTERMS.fieldType},
            'field_type_code': {"rdf_predicate": HSTERMS.fieldTypeCode},
            'field_width': {"rdf_predicate": HSTERMS.fieldWidth},
            'field_precision': {"rdf_predicate": HSTERMS.fieldPrecision},
        }


class GeometryInformationInRDF(GeometryInformation, RDFBaseModel):
    class Config:
        fields = {
            'feature_count': {"rdf_predicate": HSTERMS.featureCount},
            'geometry_type': {"rdf_predicate": HSTERMS.geometryType},
        }


class VariableInRDF(Variable, RDFBaseModel):
    class Config:
        fields = {
            'name': {"rdf_predicate": HSTERMS.name},
            'unit': {"rdf_predicate": HSTERMS.unit},
            'type': {"rdf_predicate": HSTERMS.type},
            'shape': {"rdf_predicate": HSTERMS.shape},
            'descriptive_name': {"rdf_predicate": HSTERMS.descriptive_name},
            'method': {"rdf_predicate": HSTERMS.method},
            'missing_value': {"rdf_predicate": HSTERMS.missing_value},
        }


class PublisherInRDF(Publisher, RDFBaseModel):
    class Config:
        fields = {'name': {"rdf_predicate": HSTERMS.publisherName}, 'url': {"rdf_predicate": HSTERMS.publisherURL}}


class TimeSeriesVariableInRDF(TimeSeriesVariable, RDFBaseModel):
    class Config:
        fields = {
            'variable_code': {"rdf_predicate": HSTERMS.VariableCode},
            'variable_name': {"rdf_predicate": HSTERMS.VariableName},
            'variable_type': {"rdf_predicate": HSTERMS.VariableType},
            'no_data_value': {"rdf_predicate": HSTERMS.NoDataValue},
            'variable_definition': {"rdf_predicate": HSTERMS.VariableDefinition},
            'speciation': {"rdf_predicate": HSTERMS.Speciation},
        }


class TimeSeriesSiteInRDF(TimeSeriesSite, RDFBaseModel):
    class Config:
        fields = {
            'site_code': {"rdf_predicate": HSTERMS.SiteCode},
            'site_name': {"rdf_predicate": HSTERMS.SiteName},
            'elevation_m': {"rdf_predicate": HSTERMS.Elevation_m},
            'elevation_datum': {"rdf_predicate": HSTERMS.ElevationDatum},
            'site_type': {"rdf_predicate": HSTERMS.SiteType},
            'latitude': {"rdf_predicate": HSTERMS.Latitude},
            'longitude': {"rdf_predicate": HSTERMS.Longitude},
        }


class TimeSeriesMethodInRDF(TimeSeriesMethod, RDFBaseModel):
    class Config:
        fields = {
            'method_code': {"rdf_predicate": HSTERMS.MethodCode},
            'method_name': {"rdf_predicate": HSTERMS.MethodName},
            'method_type': {"rdf_predicate": HSTERMS.MethodType},
            'method_description': {"rdf_predicate": HSTERMS.MethodDescription},
            'method_link': {"rdf_predicate": HSTERMS.MethodLink},
        }


class ProcessingLevelInRDF(ProcessingLevel, RDFBaseModel):
    class Config:
        fields = {
            'processing_level_code': {"rdf_predicate": HSTERMS.ProcessingLevelCode},
            'definition': {"rdf_predicate": HSTERMS.Definition},
            'explanation': {"rdf_predicate": HSTERMS.Explanation},
        }


class UnitInRDF(Unit, RDFBaseModel):
    class Config:
        fields = {
            'type': {"rdf_predicate": HSTERMS.UnitsType},
            'name': {"rdf_predicate": HSTERMS.UnitsName},
            'abbreviation': {"rdf_predicate": HSTERMS.UnitsAbbreviation},
        }


class UTCOffSetInRDF(UTCOffSet, RDFBaseModel):
    class Config:
        fields = {'value': {"rdf_predicate": HSTERMS.value}}


class TimeSeriesResultInRDF(TimeSeriesResult, RDFBaseModel):

    unit: UnitInRDF = Field(rdf_predicate=HSTERMS.unit, default=None)
    site: TimeSeriesSiteInRDF = Field(rdf_predicate=HSTERMS.site)
    variable: TimeSeriesVariableInRDF = Field(rdf_predicate=HSTERMS.variable)
    method: TimeSeriesMethodInRDF = Field(rdf_predicate=HSTERMS.method)
    processing_level: ProcessingLevelInRDF = Field(rdf_predicate=HSTERMS.processingLevel)
    utc_offset: UTCOffSetInRDF = Field(rdf_predicate=HSTERMS.UTCOffSet, default=None)

    class Config:
        fields = {
            'series_id': {"rdf_predicate": HSTERMS.timeSeriesResultUUID},
            'status': {"rdf_predicate": HSTERMS.Status},
            'sample_medium': {"rdf_predicate": HSTERMS.SampleMedium},
            'value_count': {"rdf_predicate": HSTERMS.ValueCount},
            'aggregation_statistic': {"rdf_predicate": HSTERMS.AggregationStatistic},
            'series_label': {"rdf_predicate": HSTERMS.SeriesLabel},
        }

    _parse_utc_offset = root_validator(pre=True)(rdf_parse_utc_offset)



from hsmodels.schemas.enums import CoverageType, DateType, RelationType, UserIdentifierType
from hsmodels.utils import to_coverage_dict


def split_dates(cls, values):
    if "dates" in values:
        for date in values['dates']:
            if date['type'] == DateType.created:
                values["created"] = date['value']
            elif date['type'] == DateType.modified:
                values["modified"] = date['value']
            elif date['type'] == DateType.published:
                values["published"] = date['value']
        del values["dates"]
    return values


def split_coverages(cls, values):
    from hsmodels.schemas.fields import BoxCoverage, PeriodCoverage, PointCoverage

    if "coverages" in values:
        for coverage in values['coverages']:
            if coverage['type'] == CoverageType.period:
                values["period_coverage"] = PeriodCoverage(**to_coverage_dict(coverage['value']))
            elif coverage['type'] == CoverageType.box:
                values["spatial_coverage"] = BoxCoverage(**to_coverage_dict(coverage['value']))
            elif coverage['type'] == CoverageType.point:
                values["spatial_coverage"] = PointCoverage(**to_coverage_dict(coverage['value']))
        del values["coverages"]
    return values


def parse_additional_metadata(cls, values):
    if "extended_metadata" in values:
        value = values["extended_metadata"]
        if isinstance(value, list):
            parsed = {}
            for em in value:
                parsed[em['key']] = em['value']
            values["additional_metadata"] = parsed
            del values["extended_metadata"]
    return values


def parse_abstract(cls, values):
    if "description" in values:
        value = values["description"]
        if isinstance(value, dict) and "abstract" in value:
            values['abstract'] = value['abstract']
            del values['description']
    return values


def parse_utc_offset_value(cls, values):
    if "utc_offset" in values:
        value = values["utc_offset"]
        if isinstance(value, dict) and "value" in value:
            values["value"] = value["value"]
            del values["utc_offset"]
    return values


def parse_url(cls, values):
    if "rdf_subject" in values:
        value = values["rdf_subject"]
        if value:
            values["url"] = values["rdf_subject"]
            del values["rdf_subject"]
    return values


def parse_relation(cls, values):
    if "type" in values or "value" in values:
        return values
    for relation_type in RelationType:
        if relation_type.name in values and values[relation_type.name]:
            values["type"] = relation_type
            values["value"] = values[relation_type.name]
            return values


def group_user_identifiers(cls, values):
    if "identifiers" not in values:
        identifiers = {}
        for identifier in UserIdentifierType:
            if identifier.name in values and values[identifier.name]:
                identifiers[identifier] = values[identifier.name]
        values["identifiers"] = identifiers
    return values



from datetime import datetime


def to_coverage_dict(value):
    value_dict = {}
    for key_value in value.split("; "):
        k, v = key_value.split("=")
        value_dict[k] = v
    return value_dict


def to_coverage_value_string(value: dict):
    return "; ".join(
        [
            "=".join([key, val.isoformat() if isinstance(val, datetime) else str(val)])
            for key, val in value.items()
            if key != "type" and val
        ]
    )



from typing import List, Union

from pydantic import AnyUrl, Field, root_validator, validator

from hsmodels.schemas.base_models import BaseMetadata
from hsmodels.schemas.enums import AggregationType
from hsmodels.schemas.fields import (
    BandInformation,
    BoxCoverage,
    BoxSpatialReference,
    CellInformation,
    FieldInformation,
    GeometryInformation,
    MultidimensionalBoxSpatialReference,
    MultidimensionalPointSpatialReference,
    PeriodCoverage,
    PointCoverage,
    PointSpatialReference,
    Rights,
    TimeSeriesResult,
    Variable,
)
from hsmodels.schemas.rdf.validators import language_constraint
from hsmodels.schemas.root_validators import parse_abstract, parse_additional_metadata, parse_url, split_coverages
from hsmodels.schemas.validators import parse_multidimensional_spatial_reference, parse_spatial_reference


class BaseAggregationMetadata(BaseMetadata):

    url: AnyUrl = Field(
        title="Aggregation URL", description="An object containing the URL of the aggregation", allow_mutation=False
    )
    title: str = Field(
        title="Aggregation title", description="A string containing a descriptive title for the aggregation"
    )
    subjects: List[str] = Field(
        default=[],
        title="Subject keywords",
        description="A list of keyword strings expressing the topic of the aggregation",
    )
    language: str = Field(
        default="eng",
        title="Language",
        description="The 3-character string for the language in which the metadata and content are expressed",
    )
    additional_metadata: dict = Field(
        default={},
        title="Extended metadata",
        description="A list of extended metadata elements expressed as key-value pairs",
    )
    spatial_coverage: Union[PointCoverage, BoxCoverage] = Field(
        default=None,
        title="Spatial coverage",
        description="An object containing the geospatial coverage for the aggregation expressed as either a bounding box or point",
    )
    period_coverage: PeriodCoverage = Field(
        default=None,
        title="Temporal coverage",
        description="An object containing the temporal coverage for a aggregation expressed as a date range",
    )
    rights: Rights = Field(
        default=None,
        title="Rights statement",
        description="An object containing information about the rights held in and over the aggregation and the license under which a aggregation is shared",
    )

    _parse_additional_metadata = root_validator(pre=True, allow_reuse=True)(parse_additional_metadata)
    _parse_coverages = root_validator(pre=True, allow_reuse=True)(split_coverages)
    _parse_url = root_validator(pre=True, allow_reuse=True)(parse_url)
    _language_constraint = validator('language', allow_reuse=True)(language_constraint)


class GeographicRasterMetadata(BaseAggregationMetadata):
    """
    A class used to represent the metadata associated with a geographic raster aggregation

    A geographic raster aggregation consists of the multiple content files that make up a
    geographic raster dataset to which aggregation-level metadata have been added. Rasters
    may have multiple files and multiple bands and are stored in HydroShare as GeoTIFF files.
    """

    class Config:
        title = 'Geographic Raster Aggregation Metadata'

        schema_config = {'read_only': ['type', 'url']}

    type: AggregationType = Field(
        const=True,
        default=AggregationType.GeographicRasterAggregation,
        title="Aggregation type",
        description="A string expressing the aggregation type from the list of HydroShare aggregation types",
        allow_mutation=False,
    )

    band_information: BandInformation = Field(
        title="Band information",
        description="An object containing information about the bands contained in the raster dataset",
    )
    spatial_reference: Union[BoxSpatialReference, PointSpatialReference] = Field(
        default=None,
        title="Spatial reference",
        description="An object containing spatial reference information for the dataset",
    )
    cell_information: CellInformation = Field(
        title="Cell information", description="An object containing information about the raster grid cells"
    )

    _parse_spatial_reference = validator("spatial_reference", pre=True, allow_reuse=True)(parse_spatial_reference)


class GeographicFeatureMetadata(BaseAggregationMetadata):
    """
    A class used to represent the metadata associated with a geographic feature aggregation

    A geographic feature aggregation consists of the multiple content files that make up an
    ESRI shapefile containing a geographic feature dataset and to which aggregation-level
    metadata have been added.
    """

    class Config:
        title = 'Geographic Feature Aggregation Metadata'

        schema_config = {'read_only': ['type', 'url']}

    type: AggregationType = Field(
        const=True,
        default=AggregationType.GeographicFeatureAggregation,
        title="Aggregation type",
        description="A string expressing the aggregation type from the list of HydroShare aggregation types",
        allow_mutation=False,
    )

    field_information: List[FieldInformation] = Field(
        title="Field information",
        description="A list of objects containing information about the fields in the dataset attribute table",
    )
    geometry_information: GeometryInformation = Field(
        title="Geometry information",
        description="An object containing information about the geometry of the features in the dataset",
    )
    spatial_reference: Union[BoxSpatialReference, PointSpatialReference] = Field(
        default=None,
        title="Spatial reference",
        description="An object containing spatial reference information for the dataset",
    )

    _parse_spatial_reference = validator("spatial_reference", pre=True, allow_reuse=True)(parse_spatial_reference)


class MultidimensionalMetadata(BaseAggregationMetadata):
    """
    A class used to represent the metadata associated with a multidimensional space-time aggregation

    An multidimensional aggregation consists of a Network Common Data Form (NetCDF) file that
    makes up a multidimensional space-time dataset to which aggregation-level metadata have
    been added.
    """

    class Config:
        title = 'Multidimensional Aggregation Metadata'

        schema_config = {'read_only': ['type', 'url']}

    type: AggregationType = Field(
        const=True,
        default=AggregationType.MultidimensionalAggregation,
        title="Aggregation type",
        description="A string expressing the aggregation type from the list of HydroShare aggregation types",
        allow_mutation=False,
    )

    variables: List[Variable] = Field(
        title="Variables",
        description="A list containing information about the variables for which data are stored in the dataset",
    )
    spatial_reference: Union[MultidimensionalBoxSpatialReference, MultidimensionalPointSpatialReference] = Field(
        default=None,
        title="Spatial reference",
        description="An object containing spatial reference information for the dataset",
    )

    _parse_spatial_reference = validator("spatial_reference", pre=True, allow_reuse=True)(
        parse_multidimensional_spatial_reference
    )


class ReferencedTimeSeriesMetadata(BaseAggregationMetadata):
    """
    A class used to represent the metadata associated with a referenced time series aggregation

    A referenced time series aggregation consists of references to specific time series
    datasets hosted on an external web service to which aggregation-level metadata have
    been added.
    """

    class Config:
        title = 'Referenced Time Series Aggregation Metadata'

        schema_config = {'read_only': ['type', 'url']}

    type: AggregationType = Field(
        const=True,
        default=AggregationType.ReferencedTimeSeriesAggregation,
        title="Aggregation type",
        description="A string expressing the aggregation type from the list of HydroShare aggregation types",
        allow_mutation=False,
    )


class FileSetMetadata(BaseAggregationMetadata):
    """
    A class used to represent the metadata associated with a file set aggregation

    A file set aggregation consists of an arbitrary collection of files that are logically
    grouped together as an aggregation and to which aggregation-level metadata have been
    added. There may be any number of files in the aggregation, and files may be of any type.
    """

    class Config:
        title = 'File Set Aggregation Metadata'

        schema_config = {'read_only': ['type', 'url']}

    type: AggregationType = Field(
        const=True,
        default=AggregationType.FileSetAggregation,
        title="Aggregation type",
        description="A string expressing the aggregation type from the list of HydroShare aggregation types",
        allow_mutation=False,
    )


class SingleFileMetadata(BaseAggregationMetadata):
    """
    A class used to represent the metadata associated with a single file aggregation

    A single file aggregation consists of a single content file to which aggregation-level
    metadata have been added.
    """

    class Config:
        title = 'Single File Aggregation Metadata'

        schema_config = {'read_only': ['type', 'url']}

    type: AggregationType = Field(
        const=True,
        default=AggregationType.SingleFileAggregation,
        title="Aggregation type",
        description="A string expressing the aggregation type from the list of HydroShare aggregation types",
        allow_mutation=False,
    )


class TimeSeriesMetadata(BaseAggregationMetadata):
    """
    A class used to represent the metadata associated with a time series aggregation

    A time series aggregation consists of one or more time series datasets to which
    aggregation-level metadata have been added. Time series datasets in HydroShare
    consist of sequences of individual data values that are ordered in time to record
    the changing trend of a certain phenomenon. They are stored in HydroShare using
    ODM2 SQLite database files.
    """

    class Config:
        title = 'Time Series Aggregation Metadata'

        schema_config = {'read_only': ['type', 'url']}

    type: AggregationType = Field(
        const=True,
        default=AggregationType.TimeSeriesAggregation,
        title="Aggregation type",
        description="A string expressing the aggregation type from the list of HydroShare aggregation types",
        allow_mutation=False,
    )

    time_series_results: List[TimeSeriesResult] = Field(
        title="Time series results",
        description="A list of time series results contained within the time series aggregation",
    )

    abstract: str = Field(default=None, title="Abstract", description="A string containing a summary of a aggregation")

    _parse_abstract = root_validator(pre=True, allow_reuse=True)(parse_abstract)



from enum import Enum

from hsmodels.namespaces import DCTERMS, HSTERMS


class StringEnum(str, Enum):
    pass


class TermEnum(StringEnum):
    pass


class CoverageType(TermEnum):
    period = str(DCTERMS.period)
    box = str(DCTERMS.box)
    point = str(DCTERMS.point)


class SpatialReferenceType(TermEnum):
    point = str(HSTERMS.point)
    box = str(HSTERMS.box)


class MultidimensionalSpatialReferenceType(TermEnum):
    point = str(DCTERMS.point)
    box = str(DCTERMS.box)


class VariableType(StringEnum):
    Char = 'Char'  # 8-bit byte that contains uninterpreted character data
    Byte = 'Byte'  # integer(8bit)
    Short = 'Short'  # signed integer (16bit)
    Int = 'Int'  # signed integer (32bit)
    Float = 'Float'  # floating point (32bit)
    Double = 'Double'  # floating point(64bit)
    Int64 = 'Int64'  # integer(64bit)
    Unsigned_Byte = 'Unsigned Byte'
    Unsigned_Short = 'Unsigned Short'
    Unsigned_Int = 'Unsigned Int'
    Unsigned_Int64 = 'Unsigned Int64'
    String = 'String'  # variable length character string
    User_Defined_Type = 'User Defined Type'  # compound, vlen, opaque, enum
    Unknown = 'Unknown'


class UserIdentifierType(StringEnum):
    google_scholar_id = "GoogleScholarID"
    research_gate_id = "ResearchGateID"
    ORCID = "ORCID"


class RelationType(StringEnum):
    isCopiedFrom = 'The content of this resource was copied from'
    isPartOf = 'The content of this resource is part of'
    hasPart = 'Has Part'
    isExecutedBy = 'The content of this resource can be executed by'
    isCreatedBy = 'The content of this resource was created by'
    isVersionOf = 'Version Of'
    isReplacedBy = 'Replaced By'
    isDataFor = 'The content of this resource serves as the data for'
    cites = 'This resource cites'
    isDescribedBy = 'This resource is described by'


class AggregationType(StringEnum):
    SingleFileAggregation = "Generic"
    FileSetAggregation = "FileSet"
    GeographicRasterAggregation = "GeoRaster"
    MultidimensionalAggregation = "NetCDF"
    GeographicFeatureAggregation = "GeoFeature"
    ReferencedTimeSeriesAggregation = "RefTimeseries"
    TimeSeriesAggregation = "TimeSeries"


class DateType(TermEnum):
    modified = str(DCTERMS.modified)
    created = str(DCTERMS.created)
    valid = str(DCTERMS.valid)
    available = str(DCTERMS.available)
    published = str(DCTERMS.published)



from hsmodels.schemas import enums, fields
from hsmodels.utils import to_coverage_dict


def parse_spatial_reference(cls, value):
    if not value:
        return value
    if value['type'] == enums.SpatialReferenceType.box:
        return fields.BoxSpatialReference(**to_coverage_dict(value['value']))
    if value['type'] == enums.SpatialReferenceType.point:
        return fields.PointSpatialReference(**to_coverage_dict(value['value']))
    return value


def parse_multidimensional_spatial_reference(cls, value):
    if value['type'] == enums.MultidimensionalSpatialReferenceType.box:
        d = to_coverage_dict(value['value'])
        return fields.MultidimensionalBoxSpatialReference(**d)
    if value['type'] == enums.MultidimensionalSpatialReferenceType.point:
        d = to_coverage_dict(value['value'])
        return fields.MultidimensionalPointSpatialReference(**d)
    return value


def parse_identifier(cls, value):
    if isinstance(value, dict) and "hydroshare_identifier" in value:
        return value['hydroshare_identifier']
    return value


def parse_sources(cls, value):
    if len(value) > 0 and isinstance(value[0], dict):
        return [f['is_derived_from'] for f in value]
    return value


def parse_spatial_coverage(cls, value):
    # This is a workaround for form submissions that do not include type
    if isinstance(value, dict) and "type" not in value:
        if "north" in value or "east" in value:
            # it's a type point
            value["type"] = "point"
        else:
            value["type"] = "box"
    return value


def list_not_empty(cls, l):
    if len(l) == 0:
        raise ValueError("list must contain at least one entry")
    return l


def validate_user_url(value):
    """Validate that a URL is a valid URL for a hydroshare user."""
    err_message = '%s is not a valid url for hydroshare user' % value
    if value:
        url_parts = value.split('/')
        if len(url_parts) != 4:
            raise ValueError(err_message)
        if url_parts[1] != 'user':
            raise ValueError(err_message)

        try:
            user_id = int(url_parts[2])
        except ValueError:
            raise ValueError(err_message)
    return value



from hsmodels.schemas.enums import CoverageType, DateType
from hsmodels.schemas.languages_iso import languages


def parse_rdf_sources(cls, value):
    if len(value) > 0 and isinstance(value[0], str):
        return [{"is_derived_from": v} for v in value]
    return value


def rdf_parse_extended_metadata(cls, value):
    from hsmodels.schemas.rdf.fields import ExtendedMetadataInRDF

    assert isinstance(value, list)
    if len(value) > 0:
        if isinstance(value[0], ExtendedMetadataInRDF):
            return value
    return [{"key": key, "value": val} for key, val in value.items()]


def rdf_parse_identifier(cls, value):
    if isinstance(value, str):
        return {"hydroshare_identifier": value}
    return value


def language_constraint(cls, language):
    if language not in [code for code, verbose in languages]:
        raise ValueError("language '{}' must be a 3 letter iso language code".format(language))
    return language


def dates_constraint(cls, dates):
    assert len(dates) >= 2
    created = list(filter(lambda d: d.type == DateType.created, dates))
    assert len(created) == 1
    created = created[0]
    modified = list(filter(lambda d: d.type == DateType.modified, dates))
    assert len(modified) == 1
    modified = modified[0]

    assert modified.value >= created.value
    return dates


def coverages_constraint(cls, coverages):
    def one_or_none_of_type(type):
        cov = list(filter(lambda d: d.type == type, coverages))
        assert len(cov) <= 1

    one_or_none_of_type(CoverageType.point)
    one_or_none_of_type(CoverageType.period)
    one_or_none_of_type(CoverageType.box)
    return coverages


def coverages_spatial_constraint(cls, coverages):
    contains_point = any(c for c in coverages if c.type == CoverageType.point)
    contains_box = any(c for c in coverages if c.type == CoverageType.box)
    if contains_point:
        assert not contains_box, "Only one type of spatial coverage is allowed, point or box"
    return coverages


def sort_creators(cls, creators):
    if not creators:
        raise ValueError("creators list must have at least one creator")
    if isinstance(next(iter(creators)), dict):
        for index, creator in enumerate(creators):
            creator["creator_order"] = index + 1
        return creators
    return sorted(creators, key=lambda creator: creator.creator_order)



from setuptools import setup, find_packages

README = 'TODO'# (pathlib.Path(__file__).parent / "README.md").read_text()

setup(
    name='hsmodels',
    version='0.2.1',
    packages=find_packages(include=['hsmodels', 'hsmodels.*', 'hsmodels.schemas.*', 'hsmodels.schemas.rdf.*'],
                           exclude=("tests",)),
    install_requires=[
        'rdflib<6.0.0',
        'pydantic==1.8.1',
        'email-validator'
    ],
    url='https://github.com/hydroshare/hsmodels',
    license='MIT',
    author='Scott Black',
    author_email='sblack@cuahsi.org',
    description='Pydantic models for HydroShare metadata',
    python_requires='>=3.6',
    long_description=README,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],

)



from datetime import datetime
from typing import Any, Dict, Type

from pydantic import BaseModel


class BaseMetadata(BaseModel):
    class Config:
        validate_assignment = True

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model) -> None:
            if hasattr(model.Config, "schema_config"):
                schema_config = model.Config.schema_config
                if "read_only" in schema_config:
                    # set readOnly in json schema
                    for field in schema_config["read_only"]:
                        schema['properties'][field]['readOnly'] = True


class BaseCoverage(BaseMetadata):
    def __str__(self):
        return "; ".join(
            [
                "=".join([key, val.isoformat() if isinstance(val, datetime) else str(val)])
                for key, val in self.__dict__.items()
                if key != "type" and val
            ]
        )



import uuid
from typing import List

from pydantic import AnyUrl, BaseModel, Field, root_validator, validator
from rdflib.term import Identifier as RDFIdentifier

from hsmodels.namespaces import CITOTERMS, DC, DCTERMS, HSRESOURCE, HSTERMS, ORE, RDF
from hsmodels.schemas.rdf.fields import (
    AwardInfoInRDF,
    ContributorInRDF,
    CoverageInRDF,
    CreatorInRDF,
    DateInRDF,
    DescriptionInRDF,
    ExtendedMetadataInRDF,
    IdentifierInRDF,
    PublisherInRDF,
    RelationInRDF,
    RightsInRDF,
    SourceInRDF,
)
from hsmodels.schemas.rdf.root_validators import (
    parse_coverages,
    parse_rdf_dates,
    parse_rdf_extended_metadata,
    rdf_parse_description,
    rdf_parse_rdf_subject,
)
from hsmodels.schemas.rdf.validators import (
    coverages_constraint,
    coverages_spatial_constraint,
    dates_constraint,
    language_constraint,
    parse_rdf_sources,
    rdf_parse_identifier,
    sort_creators,
)


def hs_uid():
    return getattr(HSRESOURCE, uuid.uuid4().hex)


class FileMap(BaseModel):
    rdf_subject: RDFIdentifier = Field(default_factory=hs_uid)
    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=ORE.Aggregation)

    is_documented_by: AnyUrl = Field(rdf_predicate=CITOTERMS.isDocumentedBy)
    files: List[AnyUrl] = Field(rdf_predicate=ORE.aggregates)
    title: str = Field(rdf_predicate=DC.title)
    is_described_by: AnyUrl = Field(rdf_predicate=ORE.isDescribedBy)


class ResourceMap(BaseModel):
    rdf_subject: RDFIdentifier = Field(default_factory=hs_uid)
    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=ORE.ResourceMap)

    describes: FileMap = Field(rdf_predicate=ORE.describes)
    identifier: str = Field(rdf_predicate=DC.identifier, default=None)
    # modified: datetime = Field(rdf_predicate=DCTERMS.modified)
    creator: str = Field(rdf_predicate=DC.creator, default=None)


class ResourceMetadataInRDF(BaseModel):

    rdf_subject: RDFIdentifier = Field(default_factory=hs_uid)
    _parse_rdf_subject = root_validator(pre=True, allow_reuse=True)(rdf_parse_rdf_subject)

    rdf_type: AnyUrl = Field(rdf_predicate=RDF.type, const=True, default=HSTERMS.CompositeResource)

    label: str = Field(default="Composite Resource", const=True)

    title: str = Field(rdf_predicate=DC.title)
    description: DescriptionInRDF = Field(rdf_predicate=DC.description, default_factory=DescriptionInRDF)
    language: str = Field(rdf_predicate=DC.language, default='eng')
    subjects: List[str] = Field(rdf_predicate=DC.subject, default=[])
    dc_type: AnyUrl = Field(rdf_predicate=DC.type, default=HSTERMS.CompositeResource, const=True)
    identifier: IdentifierInRDF = Field(rdf_predicate=DC.identifier, cont=True)
    creators: List[CreatorInRDF] = Field(rdf_predicate=DC.creator)

    contributors: List[ContributorInRDF] = Field(rdf_predicate=DC.contributor, default=[])
    sources: List[SourceInRDF] = Field(rdf_predicate=DC.source, default=[])
    relations: List[RelationInRDF] = Field(rdf_predicate=DC.relation, default=[])
    extended_metadata: List[ExtendedMetadataInRDF] = Field(rdf_predicate=HSTERMS.extendedMetadata, default=[])
    rights: RightsInRDF = Field(rdf_predicate=DC.rights, default=None)
    dates: List[DateInRDF] = Field(rdf_predicate=DC.date)
    awards: List[AwardInfoInRDF] = Field(rdf_predicate=HSTERMS.awardInfo, default=[])
    coverages: List[CoverageInRDF] = Field(rdf_predicate=DC.coverage, default=[])
    publisher: PublisherInRDF = Field(rdf_predicate=DC.publisher, default=None)
    citation: str = Field(rdf_predicate=DCTERMS.bibliographicCitation)

    _parse_coverages = root_validator(pre=True, allow_reuse=True)(parse_coverages)
    _parse_extended_metadata = root_validator(pre=True, allow_reuse=True)(parse_rdf_extended_metadata)
    _parse_rdf_dates = root_validator(pre=True, allow_reuse=True)(parse_rdf_dates)
    _parse_description = root_validator(pre=True, allow_reuse=True)(rdf_parse_description)

    _parse_identifier = validator("identifier", pre=True, allow_reuse=True)(rdf_parse_identifier)
    _parse_rdf_sources = validator("sources", pre=True, allow_reuse=True)(parse_rdf_sources)

    _language_constraint = validator('language', allow_reuse=True)(language_constraint)
    _dates_constraint = validator('dates', allow_reuse=True)(dates_constraint)
    _coverages_constraint = validator('coverages', allow_reuse=True)(coverages_constraint)
    _coverages_spatial_constraint = validator('coverages', allow_reuse=True)(coverages_spatial_constraint)
    _sort_creators = validator("creators", pre=True)(sort_creators)
